import textwrap

import jinja2
from jinja2 import Template
from conan.tools.cmake.utils import parse_extra_variable
from conan.internal.api.install.generators import relativize_path


class ConfigTemplate2:
    """
    FooConfig.cmake
    foo-config.cmake
    """
    def __init__(self, cmakedeps, require, conanfile, full_cpp_info):
        self._cmakedeps = cmakedeps
        self._require = require
        self._conanfile = conanfile
        self._full_cpp_info = full_cpp_info

    def content(self):
        t = Template(self._template, trim_blocks=True, lstrip_blocks=True,
                     undefined=jinja2.StrictUndefined)
        return t.render(self._context)

    @property
    def filename(self):
        f = self._cmakedeps.get_cmake_filename(self._conanfile)
        return f"{f}-config.cmake" if f == f.lower() else f"{f}Config.cmake"

    @property
    def _context(self):
        f = self._cmakedeps.get_cmake_filename(self._conanfile)
        targets_include = f"{f}Targets.cmake"
        pkg_name = self._conanfile.ref.name
        build_modules_paths = self._cmakedeps.get_property("cmake_build_modules", self._conanfile,
                                                           check_type=list) or []
        # FIXME: Proper escaping of paths for CMake and relativization
        # FIXME: build_module_paths coming from last config only
        build_modules_paths = [f.replace("\\", "/") for f in build_modules_paths]
        build_modules_paths = [relativize_path(p, self._cmakedeps._conanfile,
                                               "${CMAKE_CURRENT_LIST_DIR}")
                               for p in build_modules_paths]
        components = self._cmakedeps.get_property("cmake_components", self._conanfile,
                                                  check_type=list)
        if components is None:  # Lets compute the default components names
            components = []
            # This assumes that cmake_components is only defined with not multi .libs=[lib1, lib2]
            for name in self._conanfile.cpp_info.components:
                if name.startswith("_"):  # Skip private components
                    continue
                comp_components = self._cmakedeps.get_property("cmake_components", self._conanfile,
                                                               name, check_type=list)
                if comp_components:
                    components.extend(comp_components)
                else:
                    cmakename = self._cmakedeps.get_property("cmake_target_name", self._conanfile,
                                                             name)
                    if cmakename and "::" in cmakename:  # Remove package namespace
                        cmakename = cmakename.split("::", 1)[1]
                    components.append(cmakename or name)
        components = " ".join(components) if components else ""

        result = {"filename": f,
                  "components": components,
                  "pkg_name": pkg_name,
                  "targets_include_file": targets_include,
                  "build_modules_paths": build_modules_paths}

        conf_extra_variables = self._conanfile.conf.get("tools.cmake.cmaketoolchain:extra_variables",
                                                        default={}, check_type=dict)
        dep_extra_variables = self._cmakedeps.get_property("cmake_extra_variables", self._conanfile,
                                                           check_type=dict) or {}
        # The configuration variables have precedence over the dependency ones
        extra_variables = {dep: value for dep, value in dep_extra_variables.items()
                           if dep not in conf_extra_variables}
        parsed_extra_variables = {}
        for key, value in extra_variables.items():
            parsed_extra_variables[key] = parse_extra_variable("cmake_extra_variables",
                                                               key, value)
        result["extra_variables"] = parsed_extra_variables

        result.update(self._get_legacy_vars())
        return result

    def _get_legacy_vars(self):
        # Auxiliary variables for legacy consumption and try_compile cases
        pkg_name = self._conanfile.ref.name
        prefixes = self._cmakedeps.get_property("cmake_additional_variables_prefixes",
                                                self._conanfile, check_type=list) or []

        f = self._cmakedeps.get_cmake_filename(self._conanfile)
        prefixes = [f] + prefixes
        include_dirs = definitions = libraries = None
        if not self._require.build:  # To add global variables for try_compile and legacy
            aggregated_cppinfo = self._full_cpp_info.aggregated_components()
            # FIXME: Proper escaping of paths for CMake
            incdirs = [i.replace("\\", "/") for i in aggregated_cppinfo.includedirs]
            incdirs = [relativize_path(i, self._cmakedeps._conanfile, "${CMAKE_CURRENT_LIST_DIR}")
                       for i in incdirs]
            include_dirs = ";".join(incdirs)
            definitions = ""
            root_target_name = self._cmakedeps.get_property("cmake_target_name", self._conanfile)
            libraries = root_target_name or f"{pkg_name}::{pkg_name}"

        return {"additional_variables_prefixes": prefixes,
                "version": self._conanfile.ref.version,
                "include_dirs": include_dirs,
                "definitions": definitions,
                "libraries": libraries}

    @property
    def _template(self):
        return textwrap.dedent("""\
        # Requires CMake > 3.15
        if(${CMAKE_VERSION} VERSION_LESS "3.15")
            message(FATAL_ERROR "The 'CMakeDeps' generator only works with CMake >= 3.15")
        endif()

        include(${CMAKE_CURRENT_LIST_DIR}/{{ targets_include_file }})

        get_property(isMultiConfig GLOBAL PROPERTY GENERATOR_IS_MULTI_CONFIG)
        if(NOT isMultiConfig AND NOT CMAKE_BUILD_TYPE)
           message(FATAL_ERROR "Please, set the CMAKE_BUILD_TYPE variable when calling to CMake "
                               "adding the '-DCMAKE_BUILD_TYPE=<build_type>' argument.")
        endif()

        {% if components %}
        set({{filename}}_PACKAGE_PROVIDED_COMPONENTS {{components}})
        foreach(comp {%raw%}${{%endraw%}{{filename}}_FIND_COMPONENTS})
          if(NOT ${comp} IN_LIST {{filename}}_PACKAGE_PROVIDED_COMPONENTS)
            if({%raw%}${{%endraw%}{{filename}}_FIND_REQUIRED_${comp}})
              message(STATUS "Conan: Error: '{{pkg_name}}' required COMPONENT '${comp}' not found")
              set({{filename}}_FOUND FALSE)
            endif()
          endif()
        endforeach()
        {% endif %}

        ################# Global variables for try compile and legacy ##############
        {% for prefix in additional_variables_prefixes %}
        set({{ prefix }}_VERSION_STRING "{{ version }}")
        {% if include_dirs is not none %}
        set({{ prefix }}_INCLUDE_DIRS "{{ include_dirs }}" )
        set({{ prefix }}_INCLUDE_DIR "{{ include_dirs }}" )
        {% endif %}
        {% if libraries is not none %}
        set({{ prefix }}_LIBRARIES {{ libraries }} )
        {% endif %}
        {% if definitions is not none %}
        set({{ prefix }}_DEFINITIONS {{ definitions}} )
        {% endif %}
        {% endfor %}

        # build_modules_paths comes from last configuration only
        # Some build modules in ConanCenter use try_compile variables and legacy, so this
        # include() needs to happen after the above variables are defined
        {% for build_module in build_modules_paths %}
        message(STATUS "Conan: Including build module from '{{build_module}}'")
        include("{{ build_module }}")
        {% endfor %}

        # Definition of extra CMake variables from cmake_extra_variables

        {% for key, value in extra_variables.items() %}
        set({{ key }} {{ value }})
        {% endfor %}
        """)



import json
import os
import platform
import re
import textwrap

import pytest

from conan.test.utils.mocks import ConanFileMock
from conan.tools.env.environment import environment_wrap_command
from conan.test.utils.tools import TestClient, GenConanfile


@pytest.fixture()
def client():
    openssl = textwrap.dedent(r"""
        import os
        from conan import ConanFile
        from conan.tools.files import save, chdir
        class Pkg(ConanFile):
            settings = "build_type"
            def package(self):
                with chdir(self, self.package_folder):
                    echo = "@echo off\necho MYOPENSSL={}!!".format(self.settings.build_type)
                    save(self, "bin/myopenssl.bat", echo)
                    save(self, "bin/myopenssl.sh", echo)
                    os.chmod("bin/myopenssl.sh", 0o777)
            """)

    cmake = textwrap.dedent(r"""
        import os
        from conan import ConanFile
        from conan.tools.files import save, chdir
        class Pkg(ConanFile):
            type = "application"
            settings = "build_type"
            def requirements(self):
                self.requires("openssl/1.0", run=True)
            def package(self):
                with chdir(self, self.package_folder):
                    echo = "@echo off\necho MYCMAKE={}!!".format(self.settings.build_type)
                    save(self, "mycmake.bat", echo + "\ncall myopenssl.bat")
                    save(self, "mycmake.sh", echo + "\n myopenssl.sh")
                    os.chmod("mycmake.sh", 0o777)

            def package_info(self):
                # Custom buildenv not defined by cpp_info
                self.buildenv_info.prepend_path("PATH", self.package_folder)
                self.buildenv_info.define("MYCMAKEVAR", "MYCMAKEVALUE!!")
            """)

    gtest = textwrap.dedent(r"""
        import os
        from conan import ConanFile
        from conan.tools.files import save, chdir
        class Pkg(ConanFile):
            settings = "build_type"
            def package(self):
                with chdir(self, self.package_folder):
                    echo = "@echo off\necho MYGTEST={}!!".format(self.settings.build_type)
                    save(self, "bin/mygtest.bat", echo)
                    save(self, "bin/mygtest.sh", echo)
                    os.chmod("bin/mygtest.sh", 0o777)

            def package_info(self):
                self.runenv_info.define("MYGTESTVAR",
                                        "MyGTestValue{}".format(self.settings.build_type))
            """)
    client = TestClient()
    client.save({"cmake/conanfile.py": cmake,
                 "gtest/conanfile.py": gtest,
                 "openssl/conanfile.py": openssl})

    client.run("create openssl --name=openssl --version=1.0")
    client.run("create cmake --name=mycmake --version=1.0")
    client.run("create gtest --name=mygtest --version=1.0")

    myrunner_bat = "@echo off\necho MYGTESTVAR=%MYGTESTVAR%!!\n"
    myrunner_sh = "echo MYGTESTVAR=$MYGTESTVAR!!\n"
    client.save({"myrunner.bat": myrunner_bat,
                 "myrunner.sh": myrunner_sh}, clean_first=True)
    os.chmod(os.path.join(client.current_folder, "myrunner.sh"), 0o777)
    return client


def test_conanfile_txt(client):
    # conanfile.txt -(br)-> cmake
    client.save({"conanfile.txt": "[tool_requires]\nmycmake/1.0"}, clean_first=True)
    client.run("install . -s:h build_type=Debug")

    assert "mycmake/1.0" in client.out
    assert "openssl/1.0" in client.out
    ext = "bat" if platform.system() == "Windows" else "sh"  # TODO: Decide on logic .bat vs .sh
    cmd = environment_wrap_command(ConanFileMock(), "conanbuild", client.current_folder, "mycmake.{}".format(ext))
    client.run_command(cmd)

    assert "MYCMAKE=Release!!" in client.out
    assert "MYOPENSSL=Release!!" in client.out


def test_complete(client):
    app = textwrap.dedent("""
        import platform
        from conan import ConanFile
        class Pkg(ConanFile):
            requires = "openssl/1.0"
            build_requires = "mycmake/1.0"
            settings = "os"

            def build_requirements(self):
                self.test_requires("mygtest/1.0", run=True)

            def build(self):
                mybuild_cmd = "mycmake.bat" if platform.system() == "Windows" else "mycmake.sh"
                self.run(mybuild_cmd)
                mytest_cmd = "mygtest.bat" if platform.system() == "Windows" else "mygtest.sh"
                self.run(mytest_cmd, env="conanrun")
       """)

    client.save({"conanfile.py": app})
    client.run("install . -s build_type=Debug --build=missing")
    # Run the BUILD environment
    ext = "bat" if platform.system() == "Windows" else "sh"  # TODO: Decide on logic .bat vs .sh
    cmd = environment_wrap_command(ConanFileMock(),"conanbuild", client.current_folder,
                                   cmd="mycmake.{}".format(ext))
    client.run_command(cmd)
    assert "MYCMAKE=Release!!" in client.out
    assert "MYOPENSSL=Release!!" in client.out

    # Run the RUN environment
    cmd = environment_wrap_command(ConanFileMock(),"conanrun", client.current_folder,
                                   cmd="mygtest.{ext} && .{sep}myrunner.{ext}".format(ext=ext,
                                                                                      sep=os.sep))
    client.run_command(cmd)
    assert "MYGTEST=Debug!!" in client.out
    assert "MYGTESTVAR=MyGTestValueDebug!!" in client.out

    client.run("build . -s:h build_type=Debug")
    assert "MYCMAKE=Release!!" in client.out
    assert "MYOPENSSL=Release!!" in client.out
    assert "MYGTEST=Debug!!" in client.out


def test_dependents_new_buildenv():
    client = TestClient()
    boost = textwrap.dedent("""
        from conan import ConanFile
        class Boost(ConanFile):
            def package_info(self):
                self.buildenv_info.define_path("PATH", "myboostpath")
        """)
    other = textwrap.dedent("""
        from conan import ConanFile
        class Other(ConanFile):
            def requirements(self):
                self.requires("boost/1.0")

            def package_info(self):
                self.buildenv_info.append_path("PATH", "myotherpath")
                self.buildenv_info.prepend_path("PATH", "myotherprepend")
        """)
    consumer = textwrap.dedent("""
       from conan import ConanFile
       from conan.tools.env import VirtualBuildEnv
       import os
       class Lib(ConanFile):
           requires = {}
           def generate(self):
               build_env = VirtualBuildEnv(self).vars()
               with build_env.apply():
                   self.output.info("LIB PATH %s" % os.getenv("PATH"))
       """)
    client.save({"boost/conanfile.py": boost,
                 "other/conanfile.py": other,
                 "consumer/conanfile.py": consumer.format('"boost/1.0", "other/1.0"'),
                 "profile_define": "[buildenv]\nPATH=(path)profilepath",
                 "profile_append": "[buildenv]\nPATH+=(path)profilepath",
                 "profile_prepend": "[buildenv]\nPATH=+(path)profilepath"})
    client.run("create boost --name=boost --version=1.0")
    client.run("create other --name=other --version=1.0")
    client.run("install consumer")
    result = os.pathsep.join(["myotherprepend", "myboostpath", "myotherpath"])
    assert "LIB PATH {}".format(result) in client.out

    # Now test if we declare in different order, still topological order should be respected
    client.save({"consumer/conanfile.py": consumer.format('"other/1.0", "boost/1.0"')})
    client.run("install consumer")
    assert "LIB PATH {}".format(result) in client.out

    client.run("install consumer -pr=profile_define")
    assert "LIB PATH profilepath" in client.out
    client.run("install consumer -pr=profile_append")
    result = os.pathsep.join(["myotherprepend", "myboostpath", "myotherpath", "profilepath"])
    assert "LIB PATH {}".format(result) in client.out
    client.run("install consumer -pr=profile_prepend")
    result = os.pathsep.join(["profilepath", "myotherprepend", "myboostpath", "myotherpath"])
    assert "LIB PATH {}".format(result) in client.out


def test_tool_requires_conanfile_txt():
    client = TestClient(light=True)
    client.save({"conanfile.py": GenConanfile()})

    build_req = textwrap.dedent("""
        from conan import ConanFile
        class BuildReqConan(ConanFile):
            pass
        """)

    client.save({"conanfile.py": build_req})
    client.run("export . --name=build_req --version=1.0 --user=test --channel=test")

    consumer = textwrap.dedent("""
                [tool_requires]
                build_req/1.0@test/test
            """)
    client.save({"conanfile.txt": consumer}, clean_first=True)
    client.run("install . --build=missing")
    assert "build_req/1.0@test/test: Created package" in client.out


def test_profile_override_conflict():
    client = TestClient(light=True)

    test = textwrap.dedent("""
        from conan import ConanFile
        class Lib(ConanFile):

            def requirements(self):
                self.tool_requires(self.tested_reference_str)

            def test(self):
                pass
        """)
    client.save({"conanfile.py": GenConanfile("protoc"),
                 "test_package/conanfile.py": test,
                 "profile": "[tool_requires]\nprotoc/0.1"})
    client.run("create . --version 0.1 -pr=profile")
    client.run("create . --version 0.2 -pr=profile")
    assert "protoc/0.1: Already installed!" in client.out
    assert "protoc/0.2 (test package)" in client.out
    assert "WARN: The package created was 'protoc/0.1' but the reference being tested " \
           "is 'protoc/0.2'" in client.out


def test_both_context_options_error():
    # https://github.com/conan-io/conan/issues/11385
    c = TestClient()
    pkg = textwrap.dedent("""
        from conan import ConanFile
        class Pkg(ConanFile):
            name = "pkg"
            version = "0.1"
            settings = "arch"
            options = {"neon": [True, "check", False]}
            default_options = {"neon": True}

            def config_options(self):
                if "arm" not in self.settings.arch:
                    del self.options.neon
            """)
    c.save({"pkg/conanfile.py": pkg,
            "consumer/conanfile.py": GenConanfile().with_requires("pkg/0.1")
                                                   .with_build_requires("pkg/0.1")})
    c.run("export pkg")
    c.run("install consumer -s:b arch=x86_64 -s:h arch=armv8 --build=missing")
    # This failed in Conan 1.X, but now it works
    c.assert_listed_binary({"pkg/0.1": ("a0a41a189feabff576a535d071858191b90beceb", "Build")})
    c.assert_listed_binary({"pkg/0.1": ("62e589af96a19807968167026d906e63ed4de1f5", "Build")},
                           build=True)
    assert "Finalizing install" in c.out


def test_conditional_require_context():
    """ test that we can condition on the context to define a dependency
    """
    c = TestClient(light=True)
    pkg = textwrap.dedent("""
        from conan import ConanFile
        class Pkg(ConanFile):
           name = "pkg"
           version = "0.1"
           def requirements(self):
               if self.context == "host":
                   self.requires("dep/1.0")
           """)
    c.save({"dep/conanfile.py": GenConanfile("dep", "1.0"),
            "consumer/conanfile.py": pkg})
    c.run("create dep")
    c.run("create consumer")
    c.assert_listed_require({"dep/1.0": "Cache"})
    c.run("create consumer --build-require")
    assert "dep/1.0" not in c.out


class TestBuildTrackHost:

    def test_overriden_host_but_not_build(self):
        """
        Making the ``tool_requires(..., visible=True)`` works, and allows overriding, but
        propagates the build-requirement to protobuf/protoc down the graph, and VirtualBuildEnv
        will put ``protoc`` from it in the PATH. Not a problem in majority of cases, but not the
        cleanest
        """
        c = TestClient(light=True)
        pkg = textwrap.dedent("""
            from conan import ConanFile
            class ProtoBuf(ConanFile):
                name = "pkg"
                version = "0.1"
                def requirements(self):
                    self.requires("protobuf/1.0")
                def build_requirements(self):
                    self.tool_requires("protobuf/1.0", visible=True)
            """)
        c.save({"protobuf/conanfile.py": GenConanfile("protobuf"),
                "pkg/conanfile.py": pkg,
                "app/conanfile.py": GenConanfile().with_requires("pkg/0.1")
                                                  .with_requirement("protobuf/1.1", override=True)
                                                  .with_build_requirement("protobuf/1.1",
                                                                          override=True)})
        c.run("create protobuf --version=1.0")
        c.run("create protobuf --version=1.1")
        c.run("create pkg")
        c.run("install app")
        c.assert_listed_require({"protobuf/1.1": "Cache"})
        c.assert_listed_require({"protobuf/1.1": "Cache"}, build=True)

    def test_overriden_host_version(self):
        """
        Make the tool_requires follow the regular require with the expression "<host_version>"
        """
        c = TestClient(light=True)
        pkg = textwrap.dedent("""
            from conan import ConanFile
            class ProtoBuf(ConanFile):
                name = "pkg"
                version = "0.1"
                def requirements(self):
                    self.requires("protobuf/1.0")
                def build_requirements(self):
                    self.tool_requires("protobuf/<host_version>")
            """)
        c.save({"protobuf/conanfile.py": GenConanfile("protobuf"),
                "pkg/conanfile.py": pkg,
                "app/conanfile.py": GenConanfile().with_requires("pkg/0.1")
                                                  .with_requirement("protobuf/1.1", override=True)})
        c.run("create protobuf --version=1.0")
        c.run("create protobuf --version=1.1")
        c.run("create pkg")
        c.run("install pkg")  # make sure it doesn't crash
        c.run("install app")
        c.assert_listed_require({"protobuf/1.1": "Cache"})
        c.assert_listed_require({"protobuf/1.1": "Cache"}, build=True)
        # verify locks work
        c.run("lock create app")
        lock = json.loads(c.load("app/conan.lock"))
        build_requires = lock["build_requires"]
        assert len(build_requires) == 1
        assert "protobuf/1.1" in build_requires[0]
        # lock can be used
        c.run("install app --lockfile=app/conan.lock")
        c.assert_listed_require({"protobuf/1.1": "Cache"}, build=True)

    def test_overriden_host_version_version_range(self):
        """
        same as above, but using version ranges instead of overrides
        """
        c = TestClient(light=True)
        pkg = textwrap.dedent("""
            from conan import ConanFile
            class ProtoBuf(ConanFile):
                name = "pkg"
                version = "0.1"
                def requirements(self):
                    self.requires("protobuf/[*]")
                def build_requirements(self):
                    self.tool_requires("protobuf/<host_version>")
            """)
        c.save({"protobuf/conanfile.py": GenConanfile("protobuf"),
                "pkg/conanfile.py": pkg,
                "app/conanfile.py": GenConanfile().with_requires("pkg/0.1")})
        c.run("create protobuf --version=1.0")
        c.run("create pkg")
        c.run("install pkg")  # make sure it doesn't crash
        c.run("install app")
        c.assert_listed_require({"protobuf/1.0": "Cache"})
        c.assert_listed_require({"protobuf/1.0": "Cache"}, build=True)

        c.run("create protobuf --version=1.1")
        c.run("install pkg")  # make sure it doesn't crash
        c.run("install app")
        c.assert_listed_require({"protobuf/1.1": "Cache"})
        c.assert_listed_require({"protobuf/1.1": "Cache"}, build=True)
        # verify locks work
        c.run("lock create app")
        lock = json.loads(c.load("app/conan.lock"))
        build_requires = lock["build_requires"]
        assert len(build_requires) == 1
        assert "protobuf/1.1" in build_requires[0]
        # lock can be used
        c.run("install app --lockfile=app/conan.lock")
        c.assert_listed_require({"protobuf/1.1": "Cache"}, build=True)

    def test_track_host_error_nothost(self):
        """
        if no host requirement is defined, it will be an error
        """
        c = TestClient(light=True)
        c.save({"conanfile.py": GenConanfile().with_build_requirement("protobuf/<host_version>")})
        c.run("install .", assert_error=True)
        assert "ERROR:  require 'protobuf/<host_version>': " \
               "didn't find a matching host dependency" in c.out

    def test_track_host_errors_trait(self):
        """
        It is not possible to make host_version visible too
        """
        c = TestClient(light=True)
        pkg = textwrap.dedent("""
            from conan import ConanFile
            class ProtoBuf(ConanFile):
                name = "protobuf"
                def requirements(self):
                   self.tool_requires("other/<host_version>", visible=True)
            """)
        c.save({"pkg/conanfile.py": pkg})
        c.run("install pkg", assert_error=True)
        assert "ERROR: protobuf/None require 'other/<host_version>': 'host_version' " \
               "can only be used for non-visible tool_requires" in c.out

    def test_track_host_error_wrong_context(self):
        """
        it can only be used by tool_requires, not regular requires
        """
        c = TestClient(light=True)
        c.save({"conanfile.py": GenConanfile("pkg").with_requirement("protobuf/<host_version>")})
        c.run(f"install .", assert_error=True)
        assert " 'host_version' can only be used for non-visible tool_requires" in c.out

    def test_host_version_test_package(self):
        """
        https://github.com/conan-io/conan/issues/14704
        """
        c = TestClient(light=True)
        pkg = textwrap.dedent("""
                from conan import ConanFile
                class ProtoBuf(ConanFile):
                    name = "pkg"
                    version = "0.1"
                    def requirements(self):
                        self.requires("protobuf/[>=1.0]")
                    def build_requirements(self):
                        self.tool_requires("protobuf/<host_version>")
                """)
        # regular requires test_package
        c.save({"protobuf/conanfile.py": GenConanfile("protobuf"),
                "pkg/conanfile.py": pkg,
                "pkg/test_package/conanfile.py": GenConanfile().with_test("pass")})
        c.run("create protobuf --version=1.0")
        c.run(f"create pkg")
        # works without problem

        test = textwrap.dedent("""
                from conan import ConanFile
                class Test(ConanFile):
                    test_type = "explicit"

                    def build_requirements(self):
                        self.tool_requires(self.tested_reference_str)
                    def test(self):
                        pass
                """)
        c.save({"pkg/test_package/conanfile.py": test})
        c.run("create protobuf --version=1.0")
        # This used to fail
        c.run(f"create pkg")
        c.assert_listed_binary({"protobuf/1.0": ("da39a3ee5e6b4b0d3255bfef95601890afd80709",
                                                 "Cache")})
        c.assert_listed_binary({"protobuf/1.0": ("da39a3ee5e6b4b0d3255bfef95601890afd80709",
                                                 "Cache")}, build=True)

    def test_overriden_host_version_transitive_deps(self):
        """
        Make the tool_requires follow the regular require with the expression "<host_version>"
        for a transitive_deps
        """
        c = TestClient(light=True)
        c.save({"protobuf/conanfile.py": GenConanfile("protobuf"),
                "pkg/conanfile.py": GenConanfile("pkg", "0.1").with_requirement("protobuf/[>=1.0]"),
                "app/conanfile.py": GenConanfile().with_requires("pkg/0.1")
                                                  .with_tool_requirement("protobuf/<host_version>")})
        c.run("create protobuf --version=1.0")
        c.run("create protobuf --version=1.1")
        c.run("create pkg")
        c.run("install pkg")  # make sure it doesn't crash
        c.run("install app")
        c.assert_listed_require({"protobuf/1.1": "Cache"})
        c.assert_listed_require({"protobuf/1.1": "Cache"}, build=True)
        # verify locks work
        c.run("lock create app")
        lock = json.loads(c.load("app/conan.lock"))
        build_requires = lock["build_requires"]
        assert len(build_requires) == 1
        assert "protobuf/1.1" in build_requires[0]
        # lock can be used
        c.run("install app --lockfile=app/conan.lock")
        c.assert_listed_require({"protobuf/1.1": "Cache"}, build=True)

    @pytest.mark.parametrize("host_version, assert_error, assert_msg", [
        ("libgettext>", False, "gettext/0.2#d9f9eaeac9b6e403b271f04e04149df2"),
        # Error cases, just checking that we fail gracefully - no tracebacks
        ("libgettext", True, "Package 'gettext/<host_version:libgettext' not resolved"),
        (":>", True, "app/1.0 require ':/<host_version::>': didn't find a matching host dependency"),
        (">", True, "app/1.0 require '/<host_version:>': didn't find a matching host dependency"),
        (":", True, " Package 'gettext/<host_version::' not resolved"),
        ("", True, "Package 'gettext/<host_version:' not resolved: No remote defined")
    ])
    def test_host_version_different_ref(self, host_version, assert_error, assert_msg):
        tc = TestClient(light=True)
        tc.save({"gettext/conanfile.py": GenConanfile("gettext"),
                 "libgettext/conanfile.py": GenConanfile("libgettext"),
                 "app/conanfile.py": GenConanfile("app", "1.0").with_requires("libgettext/[>0.1]")
                                                   .with_tool_requirement(f"gettext/<host_version:{host_version}")})
        tc.run("create libgettext --version=0.2")
        tc.run("create gettext --version=0.1 --build-require")
        tc.run("create gettext --version=0.2 --build-require")

        tc.run("create app", assert_error=assert_error)
        assert assert_msg in tc.out

    @pytest.mark.parametrize("requires_tag,tool_requires_tag,fails", [
        ("user/channel", "user/channel", False),
        ("", "user/channel", True),
        ("auser/achannel", "anotheruser/anotherchannel", True),
    ])
    def test_overriden_host_version_user_channel(self, requires_tag, tool_requires_tag, fails):
        """
        Make the tool_requires follow the regular require with the expression "<host_version>"
        """
        c = TestClient(light=True)
        pkg = textwrap.dedent(f"""
            from conan import ConanFile
            class ProtoBuf(ConanFile):
                name = "pkg"
                version = "0.1"
                def requirements(self):
                    self.requires("protobuf/1.0@{requires_tag}")
                def build_requirements(self):
                    self.tool_requires("protobuf/<host_version>@{tool_requires_tag}")
            """)
        c.save({"protobuf/conanfile.py": GenConanfile("protobuf"),
                "pkg/conanfile.py": pkg})
        if "/" in requires_tag:
            user, channel = requires_tag.split("/", 1)
            user_channel = f"--user={user} --channel={channel}"
        else:
            user_channel = ""
        c.run(f"create protobuf --version=1.0 {user_channel}")

        c.run("create pkg", assert_error=fails)
        if fails:
            assert f"pkg/0.1 require 'protobuf/<host_version>@{tool_requires_tag}': didn't find a " \
                   "matching host dependency" in c.out
        else:
            assert "pkg/0.1: Package '39f6a091994d2d080081ea888d75ef65c1d04c8d' created" in c.out

    @pytest.mark.parametrize("shared", [True, False])
    def test_host_version_transitive_contexts(self, shared):
        # app ---------------------------------------> protobuf (shared)
        #   \---tool-require-> grpc/<host> (shared) -> protobuf (shared)
        #    \--tool-require-------(host)----------------/
        tc = TestClient(light=True)
        tc.save({"grpc/conanfile.py": GenConanfile("grpc", "0.1").with_shared_option(shared)
                .with_requirement("protobuf/0.1"),
                 "protobuf/conanfile.py": GenConanfile("protobuf", "0.1")
                .with_shared_option(shared),
                 "conanfile.py": GenConanfile("app", "0.1").with_requires("protobuf/[*]")
                .with_tool_requirement("grpc/[*]")
                .with_tool_requirement("protobuf/<host_version>")
                 })
        tc.run("export protobuf")
        tc.run("export grpc")
        tc.run("graph info .")
        assert "Conflict between" not in tc.out

    @pytest.mark.parametrize("shared", [True, False])
    def test_host_version_transitive_contexts2(self, shared):
        # app -> grpc (shared) -> protobuf (shared)
        #  \-----------------------/
        #   \---tool-require-> grpc/<host> (shared) -> protobuf (shared)
        #    \--tool-require-------(host)----------------/
        tc = TestClient(light=True)
        tc.save({"grpc/conanfile.py": GenConanfile("grpc", "0.1").with_shared_option(shared)
                .with_requirement("protobuf/0.1"),
                 "protobuf/conanfile.py": GenConanfile("protobuf", "0.1")
                .with_shared_option(shared),
                 "conanfile.py": GenConanfile("app", "0.1").with_requires("grpc/0.1")
                .with_requires("protobuf/[*]")
                .with_tool_requirement("grpc/<host_version>")
                .with_tool_requirement("protobuf/<host_version>")
                 })
        tc.run("export protobuf")
        tc.run("export grpc")
        tc.run("graph info .")
        assert "Conflict between" not in tc.out

    def test_host_version_transitive_contexts_orphan(self):
        tc = TestClient(light=True)
        tc.save({"grpc/conanfile.py": GenConanfile("grpc", "0.1")
                .with_requirement("protobuf/0.1").with_shared_option(False),
                 "protobuf/conanfile.py": GenConanfile("protobuf", "0.1").with_shared_option(False),
                 "conanfile.py": GenConanfile("app", "0.1").with_requires("grpc/0.1")
                .with_requires("protobuf/[*]")
                .with_tool_requirement("protobuf/<host_version>")
                .with_tool_requirement("grpc/<host_version>")})
        tc.run("export protobuf")
        tc.run("export grpc")
        tc.run("graph info . -o:a=*:shared=True -f=json", redirect_stdout="graph.json")
        data = json.loads(tc.load("graph.json"))

        def _assert_no_orphan(deps_graph):
            ids = set(deps_graph["nodes"].keys())
            seen = set(deps_graph["root"].keys())
            for node in deps_graph["nodes"].values():
                seen.update(node["dependencies"].keys())
            assert not ids - seen, f"Orphan nodes found: {ids - seen}"
        _assert_no_orphan(data["graph"])

    def test_user_channel_error(self):
        lib = textwrap.dedent("""
           from conan import ConanFile
           class LibWithToolConan(ConanFile):
               name = "lib_with_tool"
               package_type = "static-library"
               settings = "os", "arch"
               # {}
           """)
        c = TestClient(default_server_user=True)
        c.save({"conanfile.py": lib.format(1)})
        c.run("create . --version 1.0.0 --user foobar")
        c.save({"conanfile.py": lib.format(2)})
        c.run("create . --version 1.1.0 --user foobar")
        rev1 = c.exported_recipe_revision()
        c.save({"conanfile.py": lib.format(3)})
        c.run("create . --version 1.1.0 --user foobar")
        rev2 = c.exported_recipe_revision()
        assert rev1 != rev2

        c.run(f"upload lib_with_tool/1.1.0@foobar#{rev2} -r=default")
        c.run(f"remove lib_with_tool/1.1.0@foobar#{rev2} -c")

        dep = textwrap.dedent("""
           from conan import ConanFile
           class DepLibConan(ConanFile):
               name = "dep_lib"

               package_type = "static-library"
               settings = "os", "arch"

               def build_requirements(self):
                   self.tool_requires("lib_with_tool/<host_version>@foobar")

               def requirements(self):
                   self.requires("lib_with_tool/1.1.0@foobar")
           """)
        c.save({"conanfile.py": dep}, clean_first=True)
        c.run("create . --version=1.0.0 --user=foobar")

        conanfile = textwrap.dedent(f"""
            [requires]
            dep_lib/1.0.0@foobar
            lib_with_tool/1.1.0@foobar#{rev2}
            """)
        c.save({"conanfile.txt": conanfile}, clean_first=True)
        c.run("install .")
        # No longer produces a conflict


def test_build_missing_build_requires():
    c = TestClient(light=True)
    c.save({"tooldep/conanfile.py": GenConanfile("tooldep", "0.1"),
            "tool/conanfile.py": GenConanfile("tool", "0.1").with_tool_requires("tooldep/0.1"),
            "pkg/conanfile.py": GenConanfile("pkg", "0.1").with_tool_requires("tool/0.1"),
            "app/conanfile.py": GenConanfile().with_requires("pkg/0.1")})
    c.run("create tooldep")
    c.run("create tool")
    c.run("create pkg")
    c.run("remove tool*:* -c")
    c.run("install app")
    assert "- Build" not in c.out
    assert re.search(r"Skipped binaries(\s*)tool/0.1, tooldep/0.1", c.out)
    c.run("install app --build=missing")
    assert "- Build" not in c.out
    assert re.search(r"Skipped binaries(\s*)tool/0.1, tooldep/0.1", c.out)


def test_requirement_in_wrong_method():
    tc = TestClient(light=True)
    tc.save({"conanfile.py": textwrap.dedent("""
        from conan import ConanFile
        class Pkg(ConanFile):
            name = "pkg"
            version = "0.1"
            def configure(self):
                self.requires("foo/1.0")
        """)})
    tc.run('create . -cc="core:warnings_as_errors=[\'*\']"', assert_error=True)
    assert ("ERROR: deprecated: Requirements should only be added in the requirements()/"
            "build_requirements() methods, not configure()/config_options(), which might "
            "raise errors in the future.") in tc.out


def test_transitive_build_scripts_error():
    # https://github.com/conan-io/conan/issues/18235
    c = TestClient(light=True)
    meta = textwrap.dedent("""
        from conan import ConanFile

        class Meta(ConanFile):
            name = "meta"
            version = "0.1"
            package_type = "build-scripts"
            def requirements(self):
                self.requires("dep/0.1")
        """)
    product = textwrap.dedent("""
        from conan import ConanFile

        class Product(ConanFile):
            name = "product"
            version = "0.1"
            package_type = "build-scripts"
            def requirements(self):
                self.requires("meta/0.1")
        """)

    c.save({"dep/conanfile.py": GenConanfile("dep", "0.1"),
            "meta/conanfile.py": meta,
            "product/conanfile.py": product})
    c.run("create dep")
    c.run("create meta")
    c.run("install product")
    # It doesn't crash


def test_transitive_build_scripts_library_error():
    # https://github.com/conan-io/conan/issues/18856
    c = TestClient(light=True)
    tool = textwrap.dedent("""
        from conan import ConanFile

        class Product(ConanFile):
            name = "tool"
            version = "0.1"
            package_type = "build-scripts"
            def requirements(self):
                self.requires("dep/0.1")
        """)

    c.save({"dep/conanfile.py": GenConanfile("dep", "0.1").with_package_type("static-library"),
            "tool/conanfile.py": tool})
    c.run("create dep")
    c.run("create tool")
    # Tool-requires, not an issue
    c.run("install --tool-requires=tool/0.1")
    # requires, it crashed with traceback, now a clear error message
    c.run("install --requires=tool/0.1", assert_error=True)
    assert ("ERROR: Package 'tool/0.1' with type 'build-scripts' cannot have "
            "a 'static-library' dependency to 'dep/0.1'") in c.out



import os
import shutil

import yaml

from conan.api.output import ConanOutput
from conan.errors import ConanException
from conan.internal.errors import scoped_traceback
from conan.internal.util.files import load, save

# Related folder
WORKSPACE_FOLDER = "conanws"
# Related files
WORKSPACE_YML = "conanws.yml"
WORKSPACE_PY = "conanws.py"


class Workspace:
    """
    The base class for all workspaces
    """

    def __init__(self, folder, conan_api):
        self.folder = folder
        self.conan_data = self._conan_load_data()
        self._conan_api = conan_api
        self.output = ConanOutput(scope=f"Workspace '{self.name()}'")

    def __getattribute__(self, item):
        # Return a protected wrapper around workspace overridable callables in order to
        # be able to have clean errors if user errors in conanws.py code
        myattr = object.__getattribute__(self, item)
        if item not in ("name", "packages", "add", "remove", "clean", "build_order"):
            return myattr

        def wrapper(*args, **kwargs):
            try:
                return myattr(*args, **kwargs)
            except ConanException:
                raise
            except Exception as e:
                m = scoped_traceback(f"Error in {item}() method", e, scope="conanws.py")
                raise ConanException(f"Workspace conanws.py file: {m}")
        return wrapper

    def name(self):
        return self.conan_data.get("name") or os.path.basename(self.folder)

    def _conan_load_data(self):
        data_path = os.path.join(self.folder, WORKSPACE_YML)
        if not os.path.exists(data_path):
            return {}
        try:
            data = yaml.safe_load(load(data_path))
        except Exception as e:
            raise ConanException("Invalid yml format at {}: {}".format(WORKSPACE_YML, e))
        return data or {}

    def add(self, ref, path, output_folder):
        if not path or not os.path.isfile(path):
            raise ConanException(f"Cannot add to workspace. File not found: {path}")
        path = self._conan_rel_path(os.path.dirname(path))
        editable = {
            "path": path,
            "ref": str(ref)
        }
        if output_folder:
            editable["output_folder"] = self._conan_rel_path(output_folder)
        packages = self.conan_data.setdefault("packages", [])
        for p in packages:
            if p["path"] == path:
                self.output.warning(f"Package {path} already exists, updating its reference")
                p["ref"] = editable["ref"]
                break
        else:
            packages.append(editable)
        save(os.path.join(self.folder, WORKSPACE_YML), yaml.dump(self.conan_data))

    def remove(self, path):
        path = self._conan_rel_path(path)
        package_found = next((package_info for package_info in self.conan_data.get("packages", [])
                              if package_info["path"].replace("\\", "/") == path), None)
        if not package_found:
            raise ConanException(f"No editable package to remove from this path: {path}")
        self.conan_data["packages"].remove(package_found)
        save(os.path.join(self.folder, WORKSPACE_YML), yaml.dump(self.conan_data))
        return path

    def clean(self):
        self.output.info("Default workspace clean: Removing the output-folder of each editable")
        for package_info in self.conan_data.get("packages", []):
            editable_label = package_info.get("ref", "") or package_info['path']
            if not package_info.get("output_folder"):
                self.output.info(f"Editable {editable_label} doesn't have an output_folder defined")
                continue
            of = os.path.join(self.folder, package_info["output_folder"])
            try:
                self.output.info(f"Removing {editable_label} output folder: {of}")
                shutil.rmtree(of)
            except OSError as e:
                self.output.warning(f"Error removing {editable_label} output folder: {str(e)}")

    def _conan_rel_path(self, path):
        if path is None:
            return None
        if not os.path.isabs(path):
            raise ConanException(f"Editable path must be absolute: {path}")
        path = os.path.relpath(path, self.folder)
        return path.replace("\\", "/")  # Normalize to unix path

    def packages(self):
        return self.conan_data.get("packages", [])

    def load_conanfile(self, conanfile_path):
        conanfile_path = os.path.join(self.folder, conanfile_path, "conanfile.py")
        from conan.internal.loader import ConanFileLoader
        from conan.internal.cache.home_paths import HomePaths
        from conan.internal.conan_app import ConanFileHelpers, CmdWrapper
        cmd_wrap = CmdWrapper(HomePaths(self._conan_api.home_folder).wrapper_path)
        helpers = ConanFileHelpers(None, cmd_wrap, self._conan_api._api_helpers.global_conf,
                                   cache=None, home_folder=self._conan_api.home_folder)
        loader = ConanFileLoader(pyreq_loader=None, conanfile_helpers=helpers)
        conanfile = loader.load_named(conanfile_path, name=None, version=None, user=None,
                                      channel=None, remotes=None, graph_lock=None)
        return conanfile

    def root_conanfile(self):  # noqa
        return None

    def build_order(self, order):  # noqa
        msg = ["Packages build order:"]
        for level in order:
            for item in level:
                msg.append(f"    {item['ref']}: {item['folder']}")
        self.output.info("\n".join(msg))



import textwrap

from conan.test.utils.tools import TestClient


class TestValidPackageIdValue:

    def test_valid(self):
        c = TestClient()
        conanfile = textwrap.dedent("""
            from conan import ConanFile
            class Pkg(ConanFile):
                options = {"shared": [False, True]}
            """)

        c.save({"conanfile.py": conanfile})

        c.run("create . --name=pkg --version=0.1", assert_error=True)
        assert "pkg/0.1: Invalid: 'options.shared' value not defined" in c.out



import os
import textwrap

from conan.api.model import RecipeReference
from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient


def test_exports():
    """ Check that exported files go to the right folder
    """
    conanfile = textwrap.dedent("""
        from conan import ConanFile

        class HelloConan(ConanFile):
            name = "hello"
            version = "0.1"
            exports = "*.h"
        """)
    c = TestClient(light=True)
    c.save({"conanfile.py": conanfile,
            "hello.h": "hello",
            "data.txt": "data"})
    c.run("create .")
    ref = RecipeReference.loads("hello/0.1")
    ref_layout = c.get_latest_ref_layout(ref)

    def assert_files(folder, files):
        assert sorted(os.listdir(folder)) == sorted(files)

    assert_files(ref_layout.source(), [])
    assert_files(ref_layout.export(), ['conanfile.py', 'conanmanifest.txt', 'hello.h'])
    assert_files(ref_layout.export_sources(), [])


def test_exports_sources():
    """ Check that exported-sources files go to the right folder AND to the source folder
        """
    conanfile = textwrap.dedent("""
        from conan import ConanFile

        class HelloConan(ConanFile):
            name = "hello"
            version = "0.1"
            exports_sources = "*.h"
        """)
    c = TestClient(light=True)
    c.save({"conanfile.py": conanfile,
            "hello.h": "hello",
            "data.txt": "data"})
    c.run("create .")
    ref = RecipeReference.loads("hello/0.1")
    ref_layout = c.get_latest_ref_layout(ref)

    def assert_files(folder, files):
        assert sorted(os.listdir(folder)) == sorted(files)

    assert_files(ref_layout.source(), ['hello.h'])
    assert_files(ref_layout.export(), ['conanfile.py', 'conanmanifest.txt', ])
    assert_files(ref_layout.export_sources(), ['hello.h'])


def test_test_package_copied():
    """The exclusion of the test_package folder have been removed so now we test that indeed is
    exported"""

    client = TestClient(light=True)
    conanfile = GenConanfile().with_exports("*").with_exports_sources("*")
    client.save({"conanfile.py": conanfile,
                 "test_package/foo.txt": "bar"})
    client.run("export . --name foo --version 1.0")
    assert "Copied 2 '.txt' file" in client.out


def test_source_changes_generate_new_revisions():
    tc = TestClient(light=True)
    tc.save({"conanfile.py": GenConanfile("lib", "1.0").with_exports_sources("file.h"),
             "file.h": "Hello World!"})

    tc.run("export .")
    exported_rev = tc.exported_recipe_revision()

    tc.save({"file.h": "Bye World!"})
    tc.run("export .")
    exported_rev_new = tc.exported_recipe_revision()

    assert exported_rev != exported_rev_new



import json
import os
import re
import textwrap
import time
from collections import OrderedDict
from unittest.mock import patch, Mock

import pytest

from conan.errors import ConanException
from conan.internal.errors import ConanConnectionError
from conan.internal.util.files import save
from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.env import environment_update
from conan.test.utils.tools import TestClient, TestServer, NO_SETTINGS_PACKAGE_ID


class TestParamErrors:

    def test_default_pattern(self):
        c = TestClient()
        c.run("list")
        assert "Found 0 pkg/version" in c.out

        c.run("list -c")
        assert "Found 0 pkg/version" in c.out

        c.run('list -r="*"', assert_error=True)
        assert "ERROR: Remotes for pattern '*' can't be found" in c.out

        c.run("list --remote remote1 --cache", assert_error=True)
        assert "ERROR: Remote 'remote1' can't be found or is disabled" in c.out

    def test_query_param(self):
        c = TestClient()
        c.run("list * --graph=myjson", assert_error=True)
        assert "ERROR: Cannot define both the pattern and the graph json file" in c.out

        c.run("list * --graph-binaries=x", assert_error=True)
        assert "ERROR: --graph-recipes and --graph-binaries require a --graph input" in c.out

        c.run("list * --graph-recipes=x", assert_error=True)
        assert "ERROR: --graph-recipes and --graph-binaries require a --graph input" in c.out

        c.run("list * -p os=Linux", assert_error=True)
        assert "--package-query and --filter-xxx can only be done for binaries" in c.out

    def test_graph_file_error(self):
        # This can happen when reusing the same file in input and output
        c = TestClient(light=True)
        c.run("list --graph=graph.json", assert_error=True)
        assert "ERROR: Graph file not found" in c.out
        c.save({"graph.json": ""})
        c.run("list --graph=graph.json", assert_error=True)
        assert "ERROR: Graph file invalid JSON:" in c.out
        text = b'\x2b\x2f\x76\x38J\xe2nis\xa7'
        with open(os.path.join(c.current_folder, "graph.json"), 'wb') as handle:
            handle.write(text)
        c.run("list --graph=graph.json", assert_error=True)
        assert "ERROR: Graph file broken" in c.out

        # This can happen when using a pkg list file instead of a graph file
        c.save({"conanfile.py": GenConanfile("lib")})
        c.run("create . --version=1.0 --format=json", redirect_stdout="graph.json")
        c.run("list --graph graph.json --format=json", redirect_stdout="pkglist.json")
        c.run("list --graph pkglist.json", assert_error=True)
        assert (
            'Expected a graph file but found an unexpected JSON file format. '
            'You can create a "graph" JSON file by running'
            in c.out
        )
        assert (
            "conan [ graph-info | create | export-pkg | install | test ] --format=json > graph.json"
            in c.out
        )


@pytest.fixture(scope="module")
def client():
    servers = OrderedDict([("default", TestServer()),
                           ("other", TestServer())])
    c = TestClient(servers=servers, inputs=2*["admin", "password"])
    c.save({
        "zlib.py": GenConanfile("zlib"),
        "zlib_ng.py": GenConanfile("zlib_ng", "1.0.0"),
        "zli.py": GenConanfile("zli", "1.0.0"),
        "zli_rev2.py": GenConanfile("zli", "1.0.0").with_settings("os")
                                                   .with_package_file("f.txt", env_var="MYREV"),
        "zlix.py": GenConanfile("zlix", "1.0.0"),
        "test.py": GenConanfile("test", "1.0").with_requires("zlix/1.0.0")

                                              .with_python_requires("zlix/1.0.0"),
        "conf.py": GenConanfile("conf", "1.0")
    })
    c.run("create zli.py")
    c.run("create zlib.py --version=1.0.0 --user=user --channel=channel")
    c.run("create zlib.py --version=2.0.0 --user=user --channel=channel")
    c.run("create zlix.py")
    c.run("create test.py")
    c.run('create conf.py -c tools.info.package_id:confs="[\'tools.build:cxxflags\']"'
          ' -c tools.build:cxxflags="[\'--flag1\']"')
    c.run("upload * -r=default -c")
    c.run("upload * -r=other -c")

    time.sleep(1.0)
    # We create and upload new revisions later, to avoid timestamp overlaps (low resolution)
    with environment_update({"MYREV": "0"}):
        c.run("create zli_rev2.py -s os=Windows")
        c.run("create zli_rev2.py -s os=Linux")
    c.run("upload * -r=default -c")
    with environment_update({"MYREV": "42"}):
        c.run("create zli_rev2.py -s os=Windows")
    c.run("upload * -r=default -c")
    return c


def remove_timestamps(item):
    if isinstance(item, dict):
        if item.get("timestamp"):
            item["timestamp"] = ""
        for v in item.values():
            remove_timestamps(v)
    return item


class TestListRefs:

    @staticmethod
    def check(client, pattern, remote, expected):
        r = "-r=default" if remote else ""
        r_msg = "default" if remote else "Local Cache"
        client.run(f"list {pattern} {r}")
        expected = textwrap.indent(expected, "  ")
        expected_output = f"{r_msg}\n" + expected
        expected_output = re.sub(r"\(.*\)", "", expected_output)
        output = re.sub(r"\(.*\)", "", str(client.out))
        assert expected_output in output

    @staticmethod
    def check_json(client, pattern, remote, expected):
        r = "-r=default" if remote else ""
        r_msg = "default" if remote else "Local Cache"
        client.run(f"list {pattern} {r} --format=json", redirect_stdout="file.json")
        list_json = client.load("file.json")
        list_json = json.loads(list_json)
        assert remove_timestamps(list_json[r_msg]) == remove_timestamps(expected)

    @pytest.mark.parametrize("remote", [True, False])
    def test_list_recipes(self, client, remote):
        pattern = "z*"
        expected = textwrap.dedent(f"""\
          zli
            zli/1.0.0
          zlib
            zlib/1.0.0@user/channel
            zlib/2.0.0@user/channel
          zlix
            zlix/1.0.0
        """)
        self.check(client, pattern, remote, expected)
        expected_json = {
            "zli/1.0.0": {},
            "zlib/1.0.0@user/channel": {},
            "zlib/2.0.0@user/channel": {},
            "zlix/1.0.0": {}
        }
        self.check_json(client, pattern, remote, expected_json)

    @pytest.mark.parametrize("remote", [True, False])
    def test_list_recipes_only_user_channel(self, client, remote):
        pattern = "*@user/channel"
        expected = textwrap.dedent(f"""\
              zlib
                zlib/1.0.0@user/channel
                zlib/2.0.0@user/channel
            """)
        self.check(client, pattern, remote, expected)
        expected_json = {
            "zlib/1.0.0@user/channel": {},
            "zlib/2.0.0@user/channel": {},
        }
        self.check_json(client, pattern, remote, expected_json)

    @pytest.mark.parametrize("remote", [True, False])
    def test_list_recipes_without_user_channel(self, client, remote):
        pattern = "z*@"
        expected = textwrap.dedent(f"""\
              zli
                zli/1.0.0
              zlix
                zlix/1.0.0
            """)
        self.check(client, pattern, remote, expected)

    @pytest.mark.parametrize("remote", [True, False])
    @pytest.mark.parametrize("pattern", ["zlib", "zlib/*", "*@user/channel"])
    def test_list_recipe_versions(self, client, pattern, remote):
        expected = textwrap.dedent(f"""\
            zlib
              zlib/1.0.0@user/channel
              zlib/2.0.0@user/channel
            """)
        self.check(client, pattern, remote, expected)
        expected_json = {
            "zlib/1.0.0@user/channel": {},
            "zlib/2.0.0@user/channel": {}
        }
        self.check_json(client, pattern, remote, expected_json)

    @pytest.mark.parametrize("remote", [True, False])
    @pytest.mark.parametrize("pattern, solution", [("zlib/[*]", ("1.0.0", "2.0.0")),
                                                   ("zlib*/[*]", ("1.0.0", "2.0.0")),
                                                   ("zlib/[<2]", ("1.0.0",)),
                                                   ("zlib/[>1]", ("2.0.0",))])
    def test_list_recipe_version_ranges(self, client, pattern, solution, remote):
        expected_json = {f"zlib/{v}@user/channel": {} for v in solution}
        self.check_json(client, pattern, remote, expected_json)

    @pytest.mark.parametrize("remote", [True, False])
    def test_list_recipe_version_ranges_patterns(self, client, remote):
        pattern = "*/[>1]"
        expected_json = {'zlib/2.0.0@user/channel': {}}
        self.check_json(client, pattern, remote, expected_json)
        pattern = "z*/[<2]"
        expected_json = {'zli/1.0.0': {},
                         'zlib/1.0.0@user/channel': {},
                         'zlix/1.0.0': {}}
        self.check_json(client, pattern, remote, expected_json)

    def test_version_range_prerelease(self):
        tc = TestClient(light=True)
        tc.save({"conanfile.py": GenConanfile("foo", "1.0-pre.1")})
        tc.run("export .")
        tc.run("list * ",)
        assert "foo/1.0-pre.1" in tc.out
        tc.run("list foo/[>=1.0]")
        assert "foo/1.0-pre.1" not in tc.out
        tc.run("list foo/[>=1.0] -cc core.version_ranges:resolve_prereleases=True")
        assert "foo/1.0-pre.1" in tc.out

    @pytest.mark.parametrize("remote", [True, False])
    def test_list_recipe_versions_exact(self, client, remote):
        pattern = "zli/1.0.0"
        # by default, when a reference is complete, we show latest recipe revision
        expected = textwrap.dedent(f"""\
          zli
            zli/1.0.0
          """)
        self.check(client, pattern, remote, expected)
        expected_json = {
            "zli/1.0.0": {}
        }
        self.check_json(client, pattern, remote, expected_json)

    @pytest.mark.parametrize("remote", [True, False])
    @pytest.mark.parametrize("pattern", ["nomatch", "nomatch*", "nomatch/*"])
    def test_list_recipe_no_match(self, client, pattern, remote):
        if pattern == "nomatch":  # EXACT IS AN ERROR
            expected = "ERROR: Recipe 'nomatch' not found\n"
        else:
            expected = "WARN: There are no matching recipe references\n"

        self.check(client, pattern, remote, expected)
        if pattern == "nomatch":
            expected_json = {"error": "Recipe 'nomatch' not found"}
        else:
            expected_json = {}
        self.check_json(client, pattern, remote, expected_json)

    @pytest.mark.parametrize("remote", [True, False])
    @pytest.mark.parametrize("pattern", ["zli/1.0.0#latest",
                                         "zli/1.0.0#b58eeddfe2fd25ac3a105f72836b3360"])
    def test_list_recipe_latest_revision(self, client, remote, pattern):
        # by default, when a reference is complete, we show latest recipe revision
        expected = textwrap.dedent(f"""\
          zli
            zli/1.0.0
              revisions
                b58eeddfe2fd25ac3a105f72836b3360 (10-11-2023 10:13:13)
          """)
        self.check(client, pattern, remote, expected)
        expected_json = {
            "zli/1.0.0": {
                "revisions": {
                    "b58eeddfe2fd25ac3a105f72836b3360": {
                        "timestamp": "2023-01-10 00:25:32 UTC"
                    }
                }
            }
        }
        self.check_json(client, pattern, remote, expected_json)

    @pytest.mark.parametrize("remote", [True, False])
    def test_list_recipe_all_latest_revision(self, client, remote):
        # we can show the latest revision from several matches, if we add ``#latest``
        pattern = "zlib/*#latest"
        expected = textwrap.dedent(f"""\
            zlib
              zlib/1.0.0@user/channel
                revisions
                  ffd4bc45820ddb320ab224685b9ba3fb (10-11-2023 10:13:13)
              zlib/2.0.0@user/channel
                revisions
                  ffd4bc45820ddb320ab224685b9ba3fb (10-11-2023 10:13:13)
            """)
        self.check(client, pattern, remote, expected)

    @pytest.mark.parametrize("remote", [True, False])
    def test_list_recipe_several_revision(self, client, remote):
        # we can show the latest revision from several matches, if we add ``#latest``
        pattern = "zli/1.0.0#*"
        expected = textwrap.dedent(f"""\
            zli
              zli/1.0.0
                revisions
                  f034dc90894493961d92dd32a9ee3b78 (10-11-2023 10:13:13)
                  b58eeddfe2fd25ac3a105f72836b3360 (10-11-2023 10:13:13)
            """)
        self.check(client, pattern, remote, expected)

    @pytest.mark.parametrize("remote", [True, False])
    def test_list_recipe_multiple_revision(self, client, remote):
        pattern = "zli*#*"
        expected = textwrap.dedent(f"""\
            zli
              zli/1.0.0
                revisions
                  f034dc90894493961d92dd32a9ee3b78 (10-11-2023 10:13:13)
                  b58eeddfe2fd25ac3a105f72836b3360 (10-11-2023 10:13:13)
            zlib
              zlib/1.0.0@user/channel
                revisions
                  ffd4bc45820ddb320ab224685b9ba3fb (10-11-2023 10:13:13)
              zlib/2.0.0@user/channel
                revisions
                  ffd4bc45820ddb320ab224685b9ba3fb (10-11-2023 10:13:13)
            zlix
              zlix/1.0.0
                revisions
                  81f598d1d8648389bb7d0494fffb654e (10-11-2023 10:13:13)
            """)
        self.check(client, pattern, remote, expected)


class TestListPrefs:

    @staticmethod
    def check(client, pattern, remote, expected):
        r = "-r=default" if remote else ""
        r_msg = "default" if remote else "Local Cache"
        client.run(f"list {pattern} {r}")
        expected = textwrap.indent(expected, "  ")
        expected_output = f"{r_msg}\n" + expected
        expected_output = re.sub(r"\(.*\)", "", expected_output)
        output = re.sub(r"\(.*\)", "", str(client.out))
        assert expected_output in output

    @staticmethod
    def check_json(client, pattern, remote, expected):
        r = "-r=default" if remote else ""
        r_msg = "default" if remote else "Local Cache"
        client.run(f"list {pattern} {r} --format=json", redirect_stdout="file.json")
        list_json = client.load("file.json")
        list_json = json.loads(list_json)
        assert remove_timestamps(list_json[r_msg]) == remove_timestamps(expected)

    @pytest.mark.parametrize("remote", [True, False])
    def test_list_pkg_ids(self, client, remote):
        pattern = "zli/1.0.0:*"
        expected = textwrap.dedent(f"""\
          zli
            zli/1.0.0
              revisions
                b58eeddfe2fd25ac3a105f72836b3360 (10-11-2023 10:13:13)
                  packages
                    9a4eb3c8701508aa9458b1a73d0633783ecc2270
                      info
                        settings
                          os: Linux
                    ebec3dc6d7f6b907b3ada0c3d3cdc83613a2b715
                      info
                        settings
                          os: Windows
          """)
        self.check(client, pattern, remote, expected)
        expected_json = {
            "zli/1.0.0": {
                "revisions": {
                    "b58eeddfe2fd25ac3a105f72836b3360": {
                        "timestamp": "2023-01-10 16:30:27 UTC",
                        "packages": {
                            "9a4eb3c8701508aa9458b1a73d0633783ecc2270": {
                                "info": {
                                    "settings": {
                                        "os": "Linux"
                                    }
                                }
                            },
                            "ebec3dc6d7f6b907b3ada0c3d3cdc83613a2b715": {
                                "info": {
                                    "settings": {
                                        "os": "Windows"
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        self.check_json(client, pattern, remote, expected_json)

    @pytest.mark.parametrize("remote", [True, False])
    def test_list_pkg_ids_confs(self, client, remote):
        pattern = "conf/*:*"
        expected = textwrap.dedent("""\
          conf
            conf/1.0
              revisions
                e4e1703f72ed07c15d73a555ec3a2fa1 (10-11-2023 10:13:13)
                  packages
                    78c6fa29e8164ce399087ad6067c8f9e2f1c4ad0
                      info
                        conf
                          tools.build:cxxflags: ['--flag1']
          """)
        self.check(client, pattern, remote, expected)
        expected_json = {
            "conf/1.0": {
                "revisions": {
                    "e4e1703f72ed07c15d73a555ec3a2fa1": {
                        "timestamp": "2023-01-10 10:07:33 UTC",
                        "packages": {
                            "78c6fa29e8164ce399087ad6067c8f9e2f1c4ad0": {
                                "info": {
                                    "conf": {
                                        "tools.build:cxxflags": "['--flag1']"
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        self.check_json(client, pattern, remote, expected_json)

    @pytest.mark.parametrize("remote", [True, False])
    def test_list_pkg_ids_requires(self, client, remote):
        pattern = "test/*:*"
        expected = textwrap.dedent("""\
          test
            test/1.0
              revisions
                7df6048d3cb39b75618717987fb96453 (10-11-2023 10:13:13)
                  packages
                    81d0d9a6851a0208c2bb35fdb34eb156359d939b
                      info
                        requires
                          zlix/1.Y.Z
                        python_requires
                          zlix/1.0.Z
          """)
        self.check(client, pattern, remote, expected)
        expected_json = {
            "test/1.0": {
                "revisions": {
                    "7df6048d3cb39b75618717987fb96453": {
                        "timestamp": "2023-01-10 22:17:13 UTC",
                        "packages": {
                            "81d0d9a6851a0208c2bb35fdb34eb156359d939b": {
                                "info": {
                                    "requires": [
                                        "zlix/1.Y.Z"
                                    ],
                                    "python_requires": [
                                        "zlix/1.0.Z"
                                    ]
                                }
                            }
                        }
                    }
                }
            }
        }
        self.check_json(client, pattern, remote, expected_json)

    @pytest.mark.parametrize("remote", [True, False])
    def test_list_pkg_ids_all_rrevs(self, client, remote):
        pattern = "zli/1.0.0#*:*"
        expected = textwrap.dedent(f"""\
          zli
            zli/1.0.0
              revisions
                f034dc90894493961d92dd32a9ee3b78 (2023-01-10 22:19:58 UTC)
                  packages
                    da39a3ee5e6b4b0d3255bfef95601890afd80709
                      info
                b58eeddfe2fd25ac3a105f72836b3360 (2023-01-10 22:19:59 UTC)
                  packages
                    9a4eb3c8701508aa9458b1a73d0633783ecc2270
                      info
                        settings
                          os: Linux
                    ebec3dc6d7f6b907b3ada0c3d3cdc83613a2b715
                      info
                        settings
                          os: Windows
           """)
        self.check(client, pattern, remote, expected)

    @pytest.mark.parametrize("remote", [True, False])
    @pytest.mark.parametrize("version", ["1.0.0", "[>=1.0.0 <2]"])
    def test_list_latest_prevs(self, client, remote, version):
        pattern = f'"zli/{version}:*#latest"'
        expected = textwrap.dedent(f"""\
          zli
            zli/1.0.0
              revisions
                b58eeddfe2fd25ac3a105f72836b3360 (2023-01-10 22:27:34 UTC)
                  packages
                    9a4eb3c8701508aa9458b1a73d0633783ecc2270
                      revisions
                        9beff32b8c94ea0ce5a5e67dad95f525 (10-11-2023 10:13:13)
                      info
                        settings
                          os: Linux
                    ebec3dc6d7f6b907b3ada0c3d3cdc83613a2b715
                      revisions
                        d9b1e9044ee265092e81db7028ae10e0 (10-11-2023 10:13:13)
                      info
                        settings
                          os: Windows
          """)
        self.check(client, pattern, remote, expected)

    @pytest.mark.parametrize("remote", [True, False])
    def test_list_all_prevs(self, client, remote):
        pattern = "zli/1.0.0:*#*"
        # TODO: This is doing a package_id search, but not showing info
        expected = textwrap.dedent(f"""\
          zli
            zli/1.0.0
              revisions
                b58eeddfe2fd25ac3a105f72836b3360 (2023-01-10 22:41:09 UTC)
                  packages
                    9a4eb3c8701508aa9458b1a73d0633783ecc2270
                      revisions
                        9beff32b8c94ea0ce5a5e67dad95f525 (2023-01-10 22:41:09 UTC)
                      info
                        settings
                          os: Linux
                    ebec3dc6d7f6b907b3ada0c3d3cdc83613a2b715
                      revisions
                        24532a030b4fcdfed699511f6bfe35d3 (2023-01-10 22:41:09 UTC)
                        d9b1e9044ee265092e81db7028ae10e0 (2023-01-10 22:41:10 UTC)
                      info
                        settings
                          os: Windows
          """)
        self.check(client, pattern, remote, expected)

    @pytest.mark.parametrize("remote", [True, False])
    def test_list_package_id_all_prevs(self, client, remote):
        pattern = "zli/1.0.0:ebec3dc6d7f6b907b3ada0c3d3cdc83613a2b715#*"
        # TODO: We might want to improve the output, grouping PREVS for the
        #  same package_id
        expected_json = {
            "zli/1.0.0": {
                "revisions": {
                    "b58eeddfe2fd25ac3a105f72836b3360": {
                        "timestamp": "2023-01-10 22:45:49 UTC",
                        "packages": {
                            "ebec3dc6d7f6b907b3ada0c3d3cdc83613a2b715": {
                                "revisions": {
                                    "d9b1e9044ee265092e81db7028ae10e0": {
                                        "timestamp": "2023-01-10 22:45:49 UTC"
                                    },
                                    "24532a030b4fcdfed699511f6bfe35d3": {
                                        "timestamp": "2023-01-10 22:45:49 UTC"
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        self.check_json(client, pattern, remote, expected_json)
        expected = textwrap.dedent(f"""\
          zli
            zli/1.0.0
              revisions
                b58eeddfe2fd25ac3a105f72836b3360 (2023-01-10 22:41:09 UTC)
                  packages
                    ebec3dc6d7f6b907b3ada0c3d3cdc83613a2b715
                      revisions
                        24532a030b4fcdfed699511f6bfe35d3 (2023-01-10 22:41:09 UTC)
                        d9b1e9044ee265092e81db7028ae10e0 (2023-01-10 22:41:10 UTC)
          """)
        self.check(client, pattern, remote, expected)

    @pytest.mark.parametrize("remote", [True, False])
    def test_list_package_id_single(self, client, remote):
        pattern = "zli/1.0.0:ebec3dc6d7f6b907b3ada0c3d3cdc83613a2b715"
        expected = textwrap.dedent(f"""\
          zli
            zli/1.0.0
              revisions
                b58eeddfe2fd25ac3a105f72836b3360 (2023-01-10 23:13:12 UTC)
                  packages
                    ebec3dc6d7f6b907b3ada0c3d3cdc83613a2b715
                      info
                        settings
                          os: Windows
          """)
        self.check(client, pattern, remote, expected)

    @pytest.mark.parametrize("remote", [True, False])
    def test_list_missing_package_id(self, client, remote):
        pattern = "zli/1.0.0:nonexists_id"
        expected = "ERROR: Package ID 'zli/1.0.0:nonexists_id' not found\n"
        self.check(client, pattern, remote, expected)

    @pytest.mark.parametrize("remote", [True, False])
    def test_query(self, client, remote):
        pattern = "zli/1.0.0:* -p os=Linux"
        expected = textwrap.dedent(f"""\
          zli
            zli/1.0.0
              revisions
                b58eeddfe2fd25ac3a105f72836b3360 (10-11-2023 10:13:13)
                  packages
                    9a4eb3c8701508aa9458b1a73d0633783ecc2270
                      info
                        settings
                          os: Linux
          """)
        self.check(client, pattern, remote, expected)


def test_list_prefs_query_custom_settings():
    """
    Make sure query works for custom settings
    # https://github.com/conan-io/conan/issues/13071
    """
    c = TestClient(default_server_user=True)
    settings = textwrap.dedent("""\
        newsetting:
            value1:
            value2:
                subsetting: [1, 2, 3]
        """)
    save(c.paths.settings_path_user, settings)
    c.save({"conanfile.py": GenConanfile("pkg", "1.0").with_settings("newsetting")})
    c.run("create . -s newsetting=value1")
    c.run("create . -s newsetting=value2 -s newsetting.subsetting=1")
    c.run("create . -s newsetting=value2 -s newsetting.subsetting=2")
    c.run("upload * -r=default -c")
    c.run("list pkg/1.0:* -p newsetting=value1")
    assert "newsetting: value1" in c.out
    assert "newsetting: value2" not in c.out
    c.run("list pkg/1.0:* -p newsetting=value1 -r=default")
    assert "newsetting: value1" in c.out
    assert "newsetting: value2" not in c.out
    c.run('list pkg/1.0:* -p "newsetting=value2 AND newsetting.subsetting=1"')
    assert "newsetting: value2" in c.out
    assert "newsetting.subsetting: 1" in c.out
    assert "newsetting.subsetting: 2" not in c.out
    c.run('list pkg/1.0:* -p "newsetting=value2 AND newsetting.subsetting=1" -r=default')
    assert "newsetting: value2" in c.out
    assert "newsetting.subsetting: 1" in c.out
    assert "newsetting.subsetting: 2" not in c.out


def test_list_query_options():
    """
    Make sure query works for custom settings
    https://github.com/conan-io/conan/issues/13617
    """
    c = TestClient(default_server_user=True)
    c.save({"conanfile.py": GenConanfile("pkg", "1.0").with_option("myoption", [1, 2, 3])})
    c.run("create . -o myoption=1")
    c.run("create . -o myoption=2")
    c.run("create . -o myoption=3")

    c.run("list pkg/1.0:* -p options.myoption=1")
    assert "myoption: 1" in c.out
    assert "myoption: 2" not in c.out
    assert "myoption: 3" not in c.out
    c.run("list pkg/1.0:* -p options.myoption=2")
    assert "myoption: 1" not in c.out
    assert "myoption: 2" in c.out
    assert "myoption: 3" not in c.out

    c.run("upload * -r=default -c")
    c.run("list pkg/1.0:* -p options.myoption=1 -r=default")
    assert "myoption: 1" in c.out
    assert "myoption: 2" not in c.out
    assert "myoption: 3" not in c.out
    c.run("list pkg/1.0:* -p options.myoption=2 -r=default")
    assert "myoption: 1" not in c.out
    assert "myoption: 2" in c.out
    assert "myoption: 3" not in c.out


def test_list_empty_settings():
    """
    If settings are empty, do not crash
    """
    c = TestClient(default_server_user=True)
    c.save({"conanfile.py": GenConanfile("pkg", "1.0")})
    c.run("create .")

    c.run("list pkg/1.0:* -p os=Windows -f=json")
    revisions = json.loads(c.stdout)["Local Cache"]["pkg/1.0"]["revisions"]
    pkgs = revisions["a69a86bbd19ae2ef7eedc64ae645c531"]["packages"]
    assert pkgs == {}
    c.run("list pkg/1.0:* -p os=None -f=json")
    revisions = json.loads(c.stdout)["Local Cache"]["pkg/1.0"]["revisions"]
    pkgs = revisions["a69a86bbd19ae2ef7eedc64ae645c531"]["packages"]
    assert pkgs == {NO_SETTINGS_PACKAGE_ID: {"info": {}}}


class TestListNoUserChannel:
    def test_no_user_channel(self):
        c = TestClient(default_server_user=True)
        c.save({"zlib.py": GenConanfile("zlib"), })

        c.run("create zlib.py --version=1.0.0")
        c.run("create zlib.py --version=1.0.0 --user=user --channel=channel")
        c.run("upload * -r=default -c")

        c.run("list zlib/1.0.0#latest")
        assert "user/channel" not in c.out
        c.run("list zlib/1.0.0#latest -r=default")
        assert "user/channel" not in c.out

        c.run("list zlib/1.0.0:*")
        assert "user/channel" not in c.out
        c.run("list zlib/1.0.0:* -r=default")
        assert "user/channel" not in c.out


class TestListRemotes:
    """ advanced use case:
    - check multiple remotes output
    """

    def test_search_no_matching_recipes(self, client):
        expected_output = textwrap.dedent("""\
        Connecting to remote 'default' with user 'admin'
        Connecting to remote 'other' with user 'admin'
        Local Cache
          ERROR: Recipe 'whatever/0.1' not found
        default
          ERROR: Recipe 'whatever/0.1' not found
        other
          ERROR: Recipe 'whatever/0.1' not found
        """)

        client.run('list -c -r="*" whatever/0.1')
        assert expected_output == client.out

    def test_fail_if_no_configured_remotes(self):
        client = TestClient()
        client.run('list -r="*" whatever/1.0#123', assert_error=True)
        assert "ERROR: Remotes for pattern '*' can't be found or are disabled" in client.out

    @pytest.mark.parametrize("exc,output", [
        (ConanConnectionError("Review your network!"), "ERROR: Review your network!"),
        (ConanException("Boom!"), "ERROR: Boom!")
    ])
    def test_search_remote_errors_but_no_raising_exceptions(self, client, exc, output):
        with patch("conan.api.subapi.list._search_recipes", new=Mock(side_effect=exc)):
            client.run(f'list whatever/1.0 -r="*"')
        expected_output = textwrap.dedent(f"""\
            default
              {output}
            other
              {output}
            """)
        assert expected_output == client.out


class TestListHTML:
    def test_list_html(self):
        c = TestClient()
        c.save({"dep/conanfile.py": GenConanfile("dep", "1.2.3"),
                "pkg/conanfile.py": GenConanfile("pkg", "2.3.4").with_requires("dep/1.2.3")
                .with_settings("os", "arch").with_shared_option(False)})
        c.run("create dep")
        c.run("create pkg -s os=Windows -s arch=x86")
        # Revision is needed explicitly!
        c.run("list pkg/2.3.4#latest --format=html")
        assert "<!DOCTYPE html>" in c.stdout
        # TODO: The actual good html is missing

    def test_list_html_custom(self):
        """ test that tools.info.package_id:confs works, affecting the package_id and
        can be listed when we are listing packages
        """
        c = TestClient()
        c.save({'lib.py': GenConanfile("lib", "0.1")})
        c.run("create lib.py")
        template_folder = os.path.join(c.cache_folder, 'templates')
        c.save({"list_packages.html": '{{ base_template_path }}'}, path=template_folder)
        c.run("list lib/0.1#latest --format=html")
        assert template_folder in c.stdout


class TestListCompact:
    def test_list_compact(self):
        c = TestClient()
        c.save({"conanfile.py": GenConanfile("pkg", "1.0").with_settings("os", "arch")
                                                          .with_shared_option(False)})
        c.run("create . -s os=Windows -s arch=x86")
        c.run("create . -s os=Linux -s arch=armv8")
        c.run("create . -s os=Macos -s arch=armv8 -o shared=True")
        c.run("list pkg:* --format=compact")

        expected = textwrap.dedent("""\
            pkg/1.0#03591c8b22497dd74214e08b3bf2a56f:2a67a51fbf36a4ee345b2125dd2642be60ffd3ec
              settings: Macos, armv8
              options: shared=True
            pkg/1.0#03591c8b22497dd74214e08b3bf2a56f:2d46abc802bbffdf2af11591e3e452bc6149ea2b
              settings: Linux, armv8
              options: shared=False
            pkg/1.0#03591c8b22497dd74214e08b3bf2a56f:d2e97769569ac0a583d72c10a37d5ca26de7c9fa
              settings: Windows, x86
              options: shared=False
            """)
        assert textwrap.indent(expected, "      ") in c.stdout

    def test_list_compact_no_settings_no_options(self):
        c = TestClient()
        c.save({"pkg/conanfile.py": GenConanfile("pkg", "1.0").with_settings("os", "arch"),
                "other/conanfile.py": GenConanfile("other", "1.0")})
        c.run("create pkg -s os=Windows -s arch=x86")
        c.run("create other")
        c.run("list *:* --format=compact")
        expected_output = re.sub(r"%.* ", "%timestamp ",
                                 re.sub(r"\(.*\)", "(timestamp)", c.stdout))

        expected = textwrap.dedent("""\
            Local Cache
              other/1.0
                other/1.0#d3c8cc5e6d23ca8c6f0eaa6285c04cbd%timestamp (timestamp)
                  other/1.0#d3c8cc5e6d23ca8c6f0eaa6285c04cbd:da39a3ee5e6b4b0d3255bfef95601890afd80709
              pkg/1.0
                pkg/1.0#d24b74828b7681f08d8f5ba0e7fd791e%timestamp (timestamp)
                  pkg/1.0#d24b74828b7681f08d8f5ba0e7fd791e:c11e463c49652ba9c5adc62573ee49f966bd8417
                    settings: Windows, x86
            """)

        assert expected == expected_output

    @pytest.mark.parametrize("pattern", [
        "pkg/*",
        "pkg/1.0",
        "pkg/1.0#*",
        "pkg/1.0#*:*",
        "pkg/1.0#*:*#*",
        "pkg/1.0#a69a86bbd19ae2ef7eedc64ae645c531:*",
        "pkg/1.0#a69a86bbd19ae2ef7eedc64ae645c531:*",
        "pkg/1.0#a69a86bbd19ae2ef7eedc64ae645c531:*#*",
        "pkg/1.0#a69a86bbd19ae2ef7eedc64ae645c531:da39a3ee5e6b4b0d3255bfef95601890afd80709#*",
        "pkg/1.0#a69a86bbd19ae2ef7eedc64ae645c531:da39a3ee5e6b4b0d3255bfef95601890afd80709#"
        "0ba8627bd47edc3a501e8f0eb9a79e5e"
    ])
    def test_list_compact_patterns(self, pattern):
        c = TestClient(light=True)
        c.save({"pkg/conanfile.py": GenConanfile("pkg", "1.0")})
        c.run("create pkg")
        c.run(f"list {pattern} --format=compact")


class TestListBinaryFilter:

    @pytest.mark.parametrize("remote", [True, False])
    def test_list_filter(self, remote):
        r = "-r=default" if remote else ""
        c = TestClient(default_server_user=remote)
        c.save({"pkg/conanfile.py": GenConanfile("pkg", "1.0").with_settings("os", "arch")
                                                              .with_shared_option(False),
                "header/conanfile.py": GenConanfile("header", "1.0"),
                "profile_linux": "[settings]\nos=Linux",
                "profile_armv8": "[settings]\narch=armv8",
                "profile_shared": "[options]\n*:shared=True"})
        c.run("create pkg -s os=Windows -s arch=x86")
        c.run("create pkg -s os=Linux -s arch=armv8")
        c.run("create pkg -s os=Macos -s arch=armv8 -o shared=True")
        c.run("create header")
        if remote:
            c.run("upload *:* -r=default -c")
        pkg_key = "default" if remote else "Local Cache"

        c.run(f"list *:* -fp=profile_linux --format=json {r}")
        result = json.loads(c.stdout)
        header = result[pkg_key]["header/1.0"]["revisions"]["747cc49983b14bdd00df50a0671bd8b3"]
        assert header["packages"] == {"da39a3ee5e6b4b0d3255bfef95601890afd80709": {"info": {}}}
        pkg = result[pkg_key]["pkg/1.0"]["revisions"]["03591c8b22497dd74214e08b3bf2a56f"]
        assert len(pkg["packages"]) == 1
        settings = pkg["packages"]["2d46abc802bbffdf2af11591e3e452bc6149ea2b"]["info"]["settings"]
        assert settings == {"arch": "armv8", "os": "Linux"}

        # for linux + x86 only the header-only is a match
        c.run(f"list *:* -fp=profile_linux -fs=arch=x86 --format=json {r}")
        result = json.loads(c.stdout)
        header = result[pkg_key]["header/1.0"]["revisions"]["747cc49983b14bdd00df50a0671bd8b3"]
        assert header["packages"] == {"da39a3ee5e6b4b0d3255bfef95601890afd80709": {"info": {}}}
        pkg = result[pkg_key]["pkg/1.0"]["revisions"]["03591c8b22497dd74214e08b3bf2a56f"]
        assert pkg["packages"] == {}

        c.run(f"list *:* -fp=profile_armv8 --format=json {r}")
        result = json.loads(c.stdout)
        header = result[pkg_key]["header/1.0"]["revisions"]["747cc49983b14bdd00df50a0671bd8b3"]
        assert header["packages"] == {"da39a3ee5e6b4b0d3255bfef95601890afd80709": {"info": {}}}
        pkg = result[pkg_key]["pkg/1.0"]["revisions"]["03591c8b22497dd74214e08b3bf2a56f"]
        assert len(pkg["packages"]) == 2
        settings = pkg["packages"]["2d46abc802bbffdf2af11591e3e452bc6149ea2b"]["info"]["settings"]
        assert settings == {"arch": "armv8", "os": "Linux"}
        settings = pkg["packages"]["2a67a51fbf36a4ee345b2125dd2642be60ffd3ec"]["info"]["settings"]
        assert settings == {"arch": "armv8", "os": "Macos"}

        c.run(f"list *:* -fp=profile_shared --format=json {r}")
        result = json.loads(c.stdout)
        header = result[pkg_key]["header/1.0"]["revisions"]["747cc49983b14bdd00df50a0671bd8b3"]
        assert header["packages"] == {"da39a3ee5e6b4b0d3255bfef95601890afd80709": {"info": {}}}
        pkg = result[pkg_key]["pkg/1.0"]["revisions"]["03591c8b22497dd74214e08b3bf2a56f"]
        assert len(pkg["packages"]) == 1
        settings = pkg["packages"]["2a67a51fbf36a4ee345b2125dd2642be60ffd3ec"]["info"]["settings"]
        assert settings == {"arch": "armv8", "os": "Macos"}

        c.run(f"list *:* -fs os=Windows -fo *:shared=False --format=json {r}")
        result = json.loads(c.stdout)
        header = result[pkg_key]["header/1.0"]["revisions"]["747cc49983b14bdd00df50a0671bd8b3"]
        assert header["packages"] == {"da39a3ee5e6b4b0d3255bfef95601890afd80709": {"info": {}}}
        pkg = result[pkg_key]["pkg/1.0"]["revisions"]["03591c8b22497dd74214e08b3bf2a56f"]
        assert len(pkg["packages"]) == 1
        settings = pkg["packages"]["d2e97769569ac0a583d72c10a37d5ca26de7c9fa"]["info"]["settings"]
        assert settings == {"arch": "x86", "os": "Windows"}

        # &: will also match every package being listed, as if it was a consumer
        c.run(f"list *:* -fo &:shared=False --format=json {r}")
        result = json.loads(c.stdout)
        header = result[pkg_key]["header/1.0"]["revisions"]["747cc49983b14bdd00df50a0671bd8b3"]
        assert header["packages"] == {"da39a3ee5e6b4b0d3255bfef95601890afd80709": {"info": {}}}
        pkg = result[pkg_key]["pkg/1.0"]["revisions"]["03591c8b22497dd74214e08b3bf2a56f"]
        assert len(pkg["packages"]) == 2
        settings = pkg["packages"]["d2e97769569ac0a583d72c10a37d5ca26de7c9fa"]["info"]["settings"]
        assert settings == {"arch": "x86", "os": "Windows"}


def test_overlapping_versions():
    tc = TestClient(light=True)
    tc.save({"conanfile.py": GenConanfile("foo")})
    tc.run("export . --version=1.0")
    tc.run("export . --version=1.0.0")
    tc.run("list * -c -f=json", redirect_stdout="list.json")
    results = json.loads(tc.load("list.json"))
    assert len(results["Local Cache"]) == 2


def test_list_local_recipe_index():
    c = TestClient(light=True)
    c.run(f"new local_recipes_index -d name=pkg -d version=0.1 -d url='https://conan-fake-url.com' ")
    c.run(f"remote add local '{c.current_folder}'")

    c.run("list '*' -r=local")
    assert "Found 1 pkg/version recipes matching * in local" in c.out
    c.run("list '*/*' -r=local")
    assert "Found 1 pkg/version recipes matching */* in local" in c.out
    c.run("list '*/0.1' -r=local")
    assert "Found 1 pkg/version recipes matching */0.1 in local" in c.out
    c.run("list 'pkg/*' -r=local")
    assert "Found 1 pkg/version recipes matching pkg/* in local" in c.out
    c.run("list 'pkg/0.*' -r=local")
    assert "Found 1 pkg/version recipes matching pkg/0.* in local" in c.out
    c.run("list 'pkg' -r=local")
    assert "Found 1 pkg/version recipes matching pkg in local" in c.out
    c.run("list 'pkg/0.1' -r=local")
    assert "Found 1 pkg/version recipes matching pkg/0.1 in local" in c.out
    c.run("list 'foo' -r=local")
    assert "ERROR: Recipe 'foo' not found" in c.out
    c.run("list '/#@&%%&@#/' -r=local")
    assert "ERROR: Recipe '/' not found" in c.out
    c.run("list 'pkg/0.1@a' -r=local")
    assert "ERROR: Recipe 'pkg/0.1@a' not found" in c.out
    c.run("list 'pkg%0.1#a@b/c' -r=local")
    assert "ERROR: Recipe 'pkg%0.1' not found" in c.out



import os
import textwrap
from hashlib import sha256

from conan.test.utils.test_files import temp_folder
from conan.test.utils.tools import TestClient, GenConanfile


def test_storage_path():
    client = TestClient()
    client.save({"conanfile.py": GenConanfile()})
    tmp_folder = temp_folder(path_with_spaces=True)
    client.save_home({"global.conf": f"core.cache:storage_path={tmp_folder}"})
    client.run("create . --name=mypkg --version=0.1")
    assert f"mypkg/0.1: Package folder {tmp_folder}" in client.out
    assert os.path.isfile(os.path.join(tmp_folder, "cache.sqlite3"))

    client.run("cache path mypkg/0.1")
    assert tmp_folder in client.out


def test_wrong_home_error():
    client = TestClient(light=True)
    client.save_home({"global.conf": "core.cache:storage_path=//"})
    client.run("list *")
    assert "Couldn't initialize storage in" in client.out


def test_short_storage_path():
    c = TestClient()
    global_conf = textwrap.dedent("""\
        {% set h = hashlib.new("sha256", conan_home_folder.encode(),
                               usedforsecurity=False).hexdigest() %}
        core.cache:storage_path=C:/conan_{{h[:6]}}
        """)
    c.save_home({"global.conf": global_conf})
    c.run("config show *")
    myhash = sha256()
    myhash.update(c.cache_folder.replace("\\", "/").encode())
    myhash = myhash.hexdigest()[:6]
    assert f"core.cache:storage_path: C:/conan_{myhash}" in c.out



import pytest

from conan.tools.apple.apple import _to_apple_arch, apple_min_version_flag, is_apple_os
from conan.test.utils.mocks import MockSettings, ConanFileMock


class TestApple:
    @pytest.mark.parametrize("os_, version, sdk, subsystem, flag",
                             [("Macos", "10.1", "macosx", None, '-mmacosx-version-min=10.1'),
                              ("iOS", "10.1", "iphoneos", None, '-mios-version-min=10.1'),
                              ("iOS", "10.1", "iphonesimulator", None,
                               '-mios-simulator-version-min=10.1'),
                              ("watchOS", "10.1", "watchos", None, '-mwatchos-version-min=10.1'),
                              ("watchOS", "10.1", "watchsimulator", None,
                               '-mwatchos-simulator-version-min=10.1'),
                              ("tvOS", "10.1", "appletvos", None, '-mtvos-version-min=10.1'),
                              ("tvOS", "10.1", "appletvsimulator", None,
                               '-mtvos-simulator-version-min=10.1'),
                              ("Macos", "10.1", "macosx", "catalyst", '-mios-version-min=10.1'),
                              ("Solaris", "10.1", None, None, ''),
                              ("Macos", "10.1", None, None, '-mmacosx-version-min=10.1'),
                              ("Macos", None, "macosx", None, '')
                              ])
    def test_deployment_target_flag_name(self, os_, version, sdk, subsystem, flag):
        conanfile = ConanFileMock()
        settings = MockSettings({"os": os_,
                                 "os.version": version,
                                 "os.sdk": sdk,
                                 "os.subsystem": subsystem})
        conanfile.settings = settings
        assert apple_min_version_flag(conanfile) == flag

    @pytest.mark.parametrize("_os, result",
                             [("Macos", True), ("iOS", True), ("tvOS", True), ("watchOS", True),
                              ("Linux", False), ("Windows", False), ("Android", False)])
    def test_is_apple_os(self, _os, result):
        conanfile = ConanFileMock()
        settings = MockSettings({"os": _os})
        conanfile.settings = settings
        assert is_apple_os(conanfile) == result

    def test_to_apple_arch(self):
        assert _to_apple_arch('x86') == 'i386'
        assert _to_apple_arch('x86_64') == 'x86_64'
        assert _to_apple_arch('armv7') == 'armv7'
        assert _to_apple_arch('armv7s') == 'armv7s'
        assert _to_apple_arch('armv7k') == 'armv7k'
        assert _to_apple_arch('armv8') == 'arm64'
        assert _to_apple_arch('armv8.3') == 'arm64e'
        assert _to_apple_arch('armv8_32') == 'arm64_32'
        assert _to_apple_arch('mips') is None
        assert _to_apple_arch('mips', default='mips') == 'mips'



from conan.api.model import RecipeReference
from conan.api.output import ConanOutput


def ref_matches(ref, pattern, is_consumer):
    if not ref or not str(ref):
        assert is_consumer
        ref = RecipeReference.loads("*/*")  # FIXME: ugly
    if "[" in pattern:
        ConanOutput().warning(f"Pattern {pattern} contains a version range, which has no effect. "
                              f"Only '&' for consumer and '*' as wildcard are supported "
                              f"in this context.", warn_tag="risk")
    return ref.matches(pattern, is_consumer=is_consumer)



from conan.internal.api.new.cmake_lib import source_cpp, source_h, test_main

conanfile_exe = """import os
from conan import ConanFile
from conan.tools.files import copy
from conan.tools.layout import basic_layout
from conan.tools.premake import PremakeDeps, PremakeToolchain, Premake


class {{package_name}}Recipe(ConanFile):
    name = "{{name}}"
    version = "{{version}}"
    package_type = "application"

    # Optional metadata
    license = "<Put the package license here>"
    author = "<Put your name here> <And your email here>"
    url = "<Package recipe repository url here, for issues about the package>"
    description = "<Description of {{ name }} package here>"
    topics = ("<Put some tag here>", "<here>", "<and here>")

    # Binary configuration
    settings = "os", "compiler", "build_type", "arch"

    # Sources are located in the same place as this recipe, copy them to the recipe
    exports_sources = "premake5.lua", "src/*"

    def layout(self):
        basic_layout(self)

    def generate(self):
        deps = PremakeDeps(self)
        deps.generate()
        tc = PremakeToolchain(self)
        tc.generate()

    def build(self):
        premake = Premake(self)
        premake.configure()
        premake.build(workspace="{{name.capitalize()}}")

    def package(self):
        dest_bin = os.path.join(self.package_folder, "bin")
        build = os.path.join(self.build_folder, "bin")
        copy(self, "{{name}}", build, dest_bin, keep_path=False)
        copy(self, "{{name}}.exe", build, dest_bin, keep_path=False)


    {% if requires is defined -%}
    def requirements(self):
        {% for require in requires -%}
        self.requires("{{ require }}")
        {% endfor %}
    {%- endif %}

    {% if tool_requires is defined -%}
    def build_requirements(self):
        {% for require in tool_requires -%}
        self.tool_requires("{{ require }}")
        {% endfor %}
    {%- endif %}
"""

premake5 = """workspace "{{name.capitalize()}}"
   configurations { "Debug", "Release" }

project "{{name}}"
    kind "ConsoleApp"
    language "C++"
    files { "src/main.cpp", "src/{{name}}.cpp", "src/{{name}}.h" }

    filter "configurations:Debug"
       defines { "DEBUG" }
       symbols "On"

    filter "configurations:Release"
       defines { "NDEBUG" }
       optimize "On"
"""

test_conanfile_exe_v2 = """
from conan import ConanFile
from conan.tools.build import can_run


class {{package_name}}TestConan(ConanFile):
    settings = "os", "compiler", "build_type", "arch"

    def requirements(self):
        self.requires(self.tested_reference_str)

    def test(self):
        if can_run(self):
            self.run("{{name}}", env="conanrun")
"""

premake_exe_files = {
    "conanfile.py": conanfile_exe,
    "src/{{name}}.cpp": source_cpp,
    "src/{{name}}.h": source_h,
    "src/main.cpp": test_main,
    "premake5.lua": premake5,
    "test_package/conanfile.py": test_conanfile_exe_v2,
}



import platform
import textwrap
import pytest

from conan.test.utils.tools import TestClient
from conan.test.assets.sources import gen_function_cpp


@pytest.mark.skipif(platform.system() != "Linux", reason="QBS only for Linux now")
@pytest.mark.tool("qbs")
def test_qbsprofile():
    client = TestClient()

    conanfile = textwrap.dedent('''
        from conan import ConanFile
        from conan.tools.qbs import Qbs

        class Recipe(ConanFile):
            name = "app"
            version = "1.2"
            exports_sources = "*.cpp", "*.h", "*.qbs"
            settings = "os", "compiler", "build_type", "arch"
            generators = "QbsProfile"

            def layout(self):
                self.folders.generators = "generators"

            def build(self):
                qbs = Qbs(self)
                qbs.resolve()
                qbs.build()

            def package(self):
                qbs = Qbs(self)
                qbs.install()
        ''')

    qbsfile = textwrap.dedent('''
        CppApplication {
            files: ["main.cpp"]
        }
    ''')

    client.save({
        "conanfile.py": conanfile,
        "main.cpp": gen_function_cpp(name="main"),
        "app.qbs": qbsfile,
    }, clean_first=True)
    client.run("create .")

    assert "qbs resolve --settings-dir" in client.out
    assert "qbs build --settings-dir" in client.out
    assert "qbs install --settings-dir" in client.out



import textwrap

import pytest

from conan.test.utils.tools import TestClient, GenConanfile


class TestPackageIDRequirementsModes:

    @pytest.mark.parametrize("mode, accepted_version, rejected_version, pattern",
                             [("unrelated_mode", "2.0", "", ""),
                              ("patch_mode", "1.0.0.1", "1.0.1", "1.0.1"),
                              ("minor_mode", "1.0.1", "1.2", "1.2.Z"),
                              ("major_mode", "1.5", "2.0", "2.Y.Z"),
                              ("semver_mode", "1.5", "2.0", "2.Y.Z"),
                              ("full_mode", "1.0", "1.0.0.1", "1.0.0.1")])
    def test(self, mode, accepted_version, rejected_version, pattern):
        c = TestClient()
        package_id_text = f'self.info.requires["dep"].{mode}()'
        c.save({"dep/conanfile.py": GenConanfile("dep"),
                "pkg/conanfile.py": GenConanfile("pkg", "0.1").with_requires("dep/[*]")
                                                              .with_package_id(package_id_text)})
        c.run("create dep --version=1.0")
        pkgid = c.created_package_id("dep/1.0")
        c.run("create pkg")
        c.run(f"create dep --version={accepted_version}")
        c.run("install --requires=pkg/0.1")  # binary existing
        c.assert_listed_binary({f"dep/{accepted_version}": (pkgid, "Cache")})
        if rejected_version:
            c.run(f"create dep --version={rejected_version}")
            c.run("install --requires=pkg/0.1", assert_error=True)  # binary missing
            assert "ERROR: Missing prebuilt package for 'pkg/0.1'" in c.out
            assert f"dep/{pattern}" in c.out


@pytest.mark.parametrize("mode, pkg_id",
                             [("unrelated_mode", "da39a3ee5e6b4b0d3255bfef95601890afd80709"),
                              ("semver_mode", "13b9e753af3958dd1b2d4b3f935b04b8fb6b6760"),
                              ("patch_mode", "38d7a3ec6a09165ab3e5306f81c539a2e0a784bd"),
                              ("minor_mode", "a5e7ad26ccf4a5049090976846da1c6ed165cced"),
                              ("major_mode", "6ac597ffb99c3747ed78699f206dc1041537a8df"),
                              # This is equal to semver_mode for 0.X.Y.Z..
                              ("full_version_mode", "13b9e753af3958dd1b2d4b3f935b04b8fb6b6760"),
                              ("revision_mode", "ae5b9eeb74880aeb1cfa3db7f84c007a05ce3a76"),
                              ("full_mode", "d1b2a9538cd69363b4bae7e66c9f900b8f4c58bb")])
def test_modes(mode, pkg_id):
    c = TestClient(light=True)
    package_id_text = f'self.info.requires.{mode}()'
    c.save({"dep/conanfile.py": GenConanfile("dep", "0.1.1.1").with_settings("os"),
            "pkg/conanfile.py": GenConanfile("pkg", "0.1").with_requires("dep/[*]")
                                                          .with_package_id(package_id_text)})
    c.run("create dep -s os=Linux")
    c.run("create pkg -s os=Linux")
    pkgid = c.created_package_id("pkg/0.1")
    assert pkgid == pkg_id


class TestPackageIDError:

    def test_transitive_multi_mode_package_id(self):
        # https://github.com/conan-io/conan/issues/6942
        client = TestClient()
        client.save({"conanfile.py": GenConanfile()})
        client.run("export . --name=dep1 --version=1.0 --user=user --channel=testing")
        client.save({"conanfile.py": GenConanfile().with_require("dep1/1.0@user/testing")})
        client.run("export . --name=dep2 --version=1.0 --user=user --channel=testing")

        pkg_revision_mode = "self.info.requires.recipe_revision_mode()"
        client.save({"conanfile.py": GenConanfile().with_require("dep1/1.0@user/testing")
                                                   .with_package_id(pkg_revision_mode)})
        client.run("export . --name=dep3 --version=1.0 --user=user --channel=testing")

        client.save({"conanfile.py": GenConanfile().with_require("dep2/1.0@user/testing")
                                                   .with_require("dep3/1.0@user/testing")})
        client.run('create . --name=consumer --version=1.0 --user=user --channel=testing --build=*')
        assert "consumer/1.0@user/testing: Created" in client.out

    def test_transitive_multi_mode2_package_id(self):
        # https://github.com/conan-io/conan/issues/6942
        client = TestClient()
        # This is mandatory, otherwise it doesn't work
        client.save({"conanfile.py": GenConanfile()})
        client.run("export . --name=dep1 --version=1.0 --user=user --channel=testing")

        pkg_revision_mode = "self.info.requires.full_version_mode()"
        package_id_print = "self.output.info('PkgNames: %s' % sorted(self.info.requires.pkg_names))"
        client.save({"conanfile.py": GenConanfile().with_require("dep1/1.0@user/testing")
                    .with_package_id(pkg_revision_mode)
                    .with_package_id(package_id_print)})
        client.run("export . --name=dep2 --version=1.0 --user=user --channel=testing")

        consumer = textwrap.dedent("""
            from conan import ConanFile
            class Consumer(ConanFile):
                requires = "dep2/1.0@user/testing"
                def package_id(self):
                    self.output.info("PKGNAMES: %s" % sorted(self.info.requires.pkg_names))
                """)
        client.save({"conanfile.py": consumer})
        client.run('create . --name=consumer --version=1.0 --user=user --channel=testing --build=*')
        assert "dep2/1.0@user/testing: PkgNames: ['dep1']" in client.out
        assert "consumer/1.0@user/testing: PKGNAMES: ['dep1', 'dep2']" in client.out
        assert "consumer/1.0@user/testing: Created" in client.out

    def test_transitive_multi_mode_build_requires(self):
        # https://github.com/conan-io/conan/issues/6942
        client = TestClient()
        client.save({"conanfile.py": GenConanfile()})
        client.run("export . --name=dep1 --version=1.0 --user=user --channel=testing")
        client.run("create . --name=tool --version=1.0 --user=user --channel=testing")

        pkg_revision_mode = "self.info.requires.full_version_mode()"
        package_id_print = "self.output.info('PkgNames: %s' % sorted(self.info.requires.pkg_names))"
        client.save({"conanfile.py": GenConanfile().with_require("dep1/1.0@user/testing")
                    .with_tool_requires("tool/1.0@user/testing")
                    .with_package_id(pkg_revision_mode)
                    .with_package_id(package_id_print)})
        client.run("export . --name=dep2 --version=1.0 --user=user --channel=testing")

        consumer = textwrap.dedent("""
            from conan import ConanFile
            class Consumer(ConanFile):
                requires = "dep2/1.0@user/testing"
                build_requires = "tool/1.0@user/testing"
                def package_id(self):
                    self.output.info("PKGNAMES: %s" % sorted(self.info.requires.pkg_names))
                """)
        client.save({"conanfile.py": consumer})
        client.run('create . --name=consumer --version=1.0 --user=user --channel=testing --build=*')
        assert "dep2/1.0@user/testing: PkgNames: ['dep1']" in client.out
        assert "consumer/1.0@user/testing: PKGNAMES: ['dep1', 'dep2']" in client.out
        assert "consumer/1.0@user/testing: Created" in client.out


class TestRequirementPackageId:
    """ defining requires(..., package_id_mode)
    """

    @pytest.mark.parametrize("mode, pattern",
                             [("patch_mode", "1.2.3"),
                              ("minor_mode", "1.2.Z"),
                              ("major_mode",  "1.Y.Z")])
    def test(self, mode, pattern):
        c = TestClient(light=True)
        pkg = GenConanfile("pkg", "0.1").with_requirement("dep/1.2.3", package_id_mode=mode)
        c.save({"dep/conanfile.py": GenConanfile("dep", "1.2.3"),
                "pkg/conanfile.py": pkg})
        c.run("create dep")
        c.run("create pkg")
        c.run("list pkg:*")
        assert f"dep/{pattern}" in c.out

    def test_wrong_mode(self):
        c = TestClient(light=True)
        pkg = GenConanfile("pkg", "0.1").with_requirement("dep/1.2.3", package_id_mode="nothing")
        c.save({"dep/conanfile.py": GenConanfile("dep", "1.2.3"),
                "pkg/conanfile.py": pkg})
        c.run("create dep")
        c.run("create pkg", assert_error=True)
        assert "ERROR: require dep/1.2.3 package_id_mode='nothing' is not a known package_id_mode" \
               in c.out

    @pytest.mark.parametrize("mode, pattern",
                             [("patch_mode", "1.2.3"),
                              ("minor_mode", "1.2.Z"),
                              ("major_mode", "1.Y.Z")])
    def test_half_diamond(self, mode, pattern):
        # pkg -> libb -> liba
        #  \----(mode)----/
        c = TestClient(light=True)
        pkg = GenConanfile("pkg", "0.1").with_requirement("liba/1.2.3", package_id_mode=mode)\
                                        .with_requirement("libb/1.2.3")
        c.save({"liba/conanfile.py": GenConanfile("liba", "1.2.3"),
                "libb/conanfile.py": GenConanfile("libb", "1.2.3").with_requires("liba/1.2.3"),
                "pkg/conanfile.py": pkg})
        c.run("create liba")
        c.run("create libb")
        c.run("create pkg")
        c.run("list pkg:*")
        assert f"liba/{pattern}" in c.out
        # reverse order also works the same
        pkg = GenConanfile("pkg", "0.1").with_requirement("libb/1.2.3")\
                                        .with_requirement("liba/1.2.3", package_id_mode=mode)
        c.save({"pkg/conanfile.py": pkg})
        c.run("create pkg")
        c.run("list pkg:*")
        assert f"liba/{pattern}" in c.out

    @pytest.mark.parametrize("mode, pattern",
                             [("patch_mode", "1.2.3"),
                              ("minor_mode", "1.2.Z"),
                              ("major_mode", "1.Y.Z")])
    def test_half_diamond_conflict(self, mode, pattern):
        # pkg -> libb --(mode)-> liba
        #  \----(mode)-----------/
        # The libb->liba mode is not propagated down to pkg, so it doesn't really conflict
        c = TestClient(light=True)
        pkg = GenConanfile("pkg", "0.1").with_requirement("liba/1.2.3", package_id_mode=mode) \
            .with_requirement("libb/1.2.3")
        libb = GenConanfile("libb", "1.2.3").with_requirement("liba/1.2.3",
                                                              package_id_mode="patch_mode")
        c.save({"liba/conanfile.py": GenConanfile("liba", "1.2.3"),
                "libb/conanfile.py": libb,
                "pkg/conanfile.py": pkg})
        c.run("create liba")
        c.run("create libb")
        c.run("create pkg")
        c.run("list pkg:*")
        assert f"liba/{pattern}" in c.out
        # reverse order also works the same
        pkg = GenConanfile("pkg", "0.1").with_requirement("libb/1.2.3") \
            .with_requirement("liba/1.2.3", package_id_mode=mode)
        c.save({"pkg/conanfile.py": pkg})
        c.run("create pkg")
        c.run("list pkg:*")
        assert f"liba/{pattern}" in c.out



import textwrap

from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient


def test_settings_not_defined_consuming():
    c = TestClient()
    conanfile = textwrap.dedent("""
        from conan import ConanFile
        class Pkg(ConanFile):
            name = "pkg"
            version = "1.0"
            settings = "os", "arch"

            def build(self):
                self.output.info(f"Building with os={self.settings.os}")
                self.output.info(f"Building with arch={self.settings.arch}")

            def package_id(self):
                del self.info.settings.os
                del self.info.settings.arch
        """)
    c.save({"conanfile.py": conanfile,
            "profile": ""})
    c.run("create . -pr=profile -s os=Windows -s arch=armv8")
    assert "Building with os=Windows" in c.out
    assert "Building with arch=armv8" in c.out
    c.run("install --requires=pkg/1.0 -pr=profile")
    # It doesn't fail, even if settings not defined
    c.run("install --requires=pkg/1.0 -pr=profile -s os=Linux -s arch=x86")
    # It doesn't fail, even if settings different value


def test_settings_undefined():
    client = TestClient()
    client.save({
        "conanfile.py": GenConanfile(name="hello", version="1.0")
    })
    # Undefined settings field
    client.run("install . -s foo=None", assert_error=True)
    assert "'settings.foo' doesn't exist for 'settings'" in client.out
    client.run("install . -s foo.bar=None", assert_error=True)
    assert "'settings.foo' doesn't exist for 'settings'" in client.out



import os
import pytest

from conan.api.model import Remote
from conan.internal.model.conf import ConfDefinition
from conan.internal.model.manifest import FileTreeManifest
from conan.api.model import PkgReference
from conan.api.model import RecipeReference
from conan.internal.paths import CONANFILE, CONANINFO, CONAN_MANIFEST
from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.env import environment_update
from conan.test.utils.server_launcher import TestServerLauncher
from conan.test.utils.test_files import temp_folder
from conan.test.utils.tools import get_free_port
from conan.internal.rest.conan_requester import ConanRequester
from conan.internal.rest.rest_client import RestApiClient
from conan.internal.util.files import md5, save


class TestRestApi:
    """Open a real server (sockets) to test rest_api function."""

    @pytest.fixture(scope="class", autouse=True)
    def setup_class(self):
        with environment_update({"CONAN_SERVER_PORT": str(get_free_port())}):
            read_perms = [("*/*@*/*", "private_user")]
            write_perms = [("*/*@*/*", "private_user")]
            self.server = TestServerLauncher(server_capabilities=['ImCool', 'TooCool'],
                                             read_permissions=read_perms,
                                             write_permissions=write_perms)
            self.server.start()

            config = ConfDefinition()
            requester = ConanRequester(config)

            remote = Remote("myremote", f"http://127.0.0.1:{self.server.port}", True, True)
            self.api = RestApiClient(remote, None, requester, config)
            self.api._token = self.api.authenticate(user="private_user", password="private_pass")
            # Necessary for setup_class approach
            TestRestApi.api = self.api
            TestRestApi.server = self.server
        yield

        self.server.stop()
        self.server.clean()

    def test_get_conan(self):
        # Upload a conans
        ref = RecipeReference.loads("conan1/1.0.0@private_user/testing#myreciperev")
        self._upload_recipe(ref)

        # Get the conans
        tmp_dir = temp_folder()
        self.api.get_recipe(ref, tmp_dir, metadata=None, only_metadata=False)
        assert CONANFILE in os.listdir(tmp_dir)
        assert CONAN_MANIFEST in os.listdir(tmp_dir)

    def test_get_package(self):
        # Upload a conans
        ref = RecipeReference.loads("conan3/1.0.0@private_user/testing#myreciperev")
        self._upload_recipe(ref)

        # Upload an package
        pref = PkgReference(ref, "1F23223EFDA2", "mypackagerev")
        self._upload_package(pref)

        # Get the package
        tmp_dir = temp_folder()
        self.api.get_package(pref, tmp_dir, metadata=None, only_metadata=False)
        # The hello.cpp file is not downloaded!
        assert "hello.cpp" not in os.listdir(tmp_dir)

    def test_upload_huge_conan(self):
        ref = RecipeReference.loads("conanhuge/1.0.0@private_user/testing#myreciperev")
        self._upload_recipe(ref, {"file9.cpp": ""})

        tmp = temp_folder()
        files = self.api.get_recipe(ref, tmp, metadata=None, only_metadata=False)
        assert files is not None
        assert not os.path.exists(os.path.join(tmp, "file9.cpp"))

    def test_search(self):
        # Upload a conan1
        conan_name1 = "HelloOnly/0.10@private_user/testing#myreciperev"
        ref1 = RecipeReference.loads(conan_name1)
        self._upload_recipe(ref1)

        # Upload a package
        conan_info = """[settings]
    arch=x86_64
    compiler=gcc
    os=Linux
[options]
    386=False
[requires]
    Hello
    Bye/2.9
    Say/2.1@user/testing
    Chat/2.1@user/testing:SHA_ABC
"""
        pref = PkgReference(ref1, "1F23223EFDA", "mypackagerev")
        self._upload_package(pref, {CONANINFO: conan_info})

        # Upload a conan2
        conan_name2 = "helloonlyToo/2.1@private_user/stable#myreciperev"
        ref2 = RecipeReference.loads(conan_name2)
        self._upload_recipe(ref2)

        # Get the info about this ConanFileReference
        info = self.api.search_packages(ref1)
        assert conan_info == info["1F23223EFDA"]["content"]

        # Search packages
        results = self.api.search("HelloOnly*", ignorecase=False)
        results = [RecipeReference(r.name, r.version, r.user, r.channel, revision=None)
                   for r in results]
        ref1.revision = None
        assert results == [ref1]

    def test_remove(self):
        # Upload a conans
        ref = RecipeReference.loads("MyFirstConan/1.0.0@private_user/testing#myreciperev")
        self._upload_recipe(ref)
        ref.revision = "myreciperev"
        path1 = self.server.server_store.base_folder(ref)
        assert os.path.exists(path1)
        # Remove conans and packages
        self.api.remove_recipe(ref)
        assert not os.path.exists(path1)

    def test_remove_packages(self):
        ref = RecipeReference.loads("MySecondConan/2.0.0@private_user/testing#myreciperev")
        self._upload_recipe(ref)

        folders = {}
        for sha in ["1", "2", "3", "4", "5"]:
            # Upload an package
            pref = PkgReference(ref, sha, "mypackagerev")
            self._upload_package(pref, {CONANINFO: ""})
            folder = self.server.server_store.package(pref)
            assert os.path.exists(folder)
            folders[sha] = folder

        data = self.api.search_packages(ref)
        assert len(data) == 5

        self.api.remove_packages([PkgReference(ref, "1")])
        assert os.path.exists(self.server.server_store.base_folder(ref))
        assert not os.path.exists(folders["1"])
        assert os.path.exists(folders["2"])
        assert os.path.exists(folders["3"])
        assert os.path.exists(folders["4"])
        assert os.path.exists(folders["5"])

        self.api.remove_packages([PkgReference(ref, "2"), PkgReference(ref, "3")])
        assert os.path.exists(self.server.server_store.base_folder(ref))
        assert not os.path.exists(folders["1"])
        assert not os.path.exists(folders["2"])
        assert not os.path.exists(folders["3"])
        assert os.path.exists(folders["4"])
        assert os.path.exists(folders["5"])

        self.api.remove_all_packages(ref)
        assert os.path.exists(self.server.server_store.base_folder(ref))
        for sha in ["1", "2", "3", "4", "5"]:
            assert not os.path.exists(folders[sha])

    def _upload_package(self, package_reference, base_files=None):

        files = {"conanfile.py": GenConanfile("3").with_requires("1", "12").with_exports("*"),
                 "hello.cpp": "hello",
                 "conanmanifest.txt": ""}
        if base_files:
            files.update(base_files)

        tmp_dir = temp_folder()
        abs_paths = {}
        for filename, content in files.items():
            abs_path = os.path.join(tmp_dir, filename)
            save(abs_path, str(content))
            abs_paths[filename] = abs_path

        self.api.upload_package(package_reference, abs_paths)

    def _upload_recipe(self, ref, base_files=None):

        files = {"conanfile.py": GenConanfile("3").with_requires("1", "12")}
        if base_files:
            files.update(base_files)
        content = """
from conan import ConanFile

class MyConan(ConanFile):
    name = "%s"
    version = "%s"
    settings = arch, compiler, os
""" % (ref.name, ref.version)
        files[CONANFILE] = content
        files_md5s = {filename: md5(content) for filename, content in files.items()}
        conan_digest = FileTreeManifest(123123123, files_md5s)

        tmp_dir = temp_folder()
        abs_paths = {}
        for filename, content in files.items():
            abs_path = os.path.join(tmp_dir, filename)
            save(abs_path, content)
            abs_paths[filename] = abs_path
        abs_paths[CONAN_MANIFEST] = os.path.join(tmp_dir, CONAN_MANIFEST)
        conan_digest.save(tmp_dir)

        self.api.upload_recipe(ref, abs_paths)



import os

import pytest

from conan.api.conan_api import ConanAPI
from conan.cli.cli import Cli, main
from conan.test.utils.env import environment_update
from conan.test.utils.mocks import RedirectedTestOutput
from conan.test.utils.test_files import temp_folder
from conan.test.utils.tools import redirect_output


def test_cli():
    """ make sure the CLi can be reused
    https://github.com/conan-io/conan/issues/14044
    """
    folder = temp_folder()
    api = ConanAPI(cache_folder=folder)
    cli = Cli(api)
    cli2 = Cli(api)

    stdout = RedirectedTestOutput()
    stderr = RedirectedTestOutput()
    with redirect_output(stderr, stdout):
        cli.run(["list", "*"])
        cli.run(["list", "*"])
        cli2.run(["list", "*"])
        cli.run(["list", "*"])

    stdout = RedirectedTestOutput()
    stderr = RedirectedTestOutput()
    with redirect_output(stderr, stdout):
        cli.run()
    # Running without args shows help, but doesn't error
    assert "Consumer commands" in stdout.getvalue()


def test_basic_api():
    api = ConanAPI(cache_folder=temp_folder())
    result = api.remotes.list()
    assert result[0].name == "conancenter"


def test_api_command():
    # The ``CommandAPI`` requires a bit more of setup
    api = ConanAPI(cache_folder=temp_folder())
    cli = Cli(api)
    cli.add_commands()
    result = api.command.run(["remote", "list"])
    assert result[0].name == "conancenter"


def test_main():
    cache_folder = os.path.join(temp_folder(), "custom")
    with environment_update({"CONAN_HOME": cache_folder}):
        with pytest.raises(SystemExit) as e:
            main(["list", "*"])
        assert e.type == SystemExit
        assert e.value.code == 0  # success



import json
import os
import textwrap

import pytest

from conan.test.utils.tools import TestClient


@pytest.fixture
def fixture_client():
    tc = TestClient(light=True, default_server_user=True)

    patch_file = textwrap.dedent("""
        diff --git a/foo.txt b/foo.txt
        new file mode 100644
        index 0000000..e311fbb
        --- /dev/null
        +++ b/foo.txt
        @@ -0,0 +1 @@
        +{version}
        """)

    conandata_yml = textwrap.dedent("""
        patches:
          "{version}":
            - patch_file: "patches/patch.patch"
        """)

    conanfile = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.files import apply_conandata_patches, export_conandata_patches, save

        class TestConan(ConanFile):
            name = "pkg"

            exports = "{version}.txt"

            def export_sources(self):
                export_conandata_patches(self)

            def source(self):
                save(self, "myfile.txt", "{version}")
                if self.version == "2.0":
                    save(self, "new-file-for-v2.txt", "a new file for the new version")
                apply_conandata_patches(self)
        """)

    tc.save({
        "v1/conanfile.py": conanfile.format(version="v1"),
        "v1/conandata.yml": conandata_yml.format(version="1.0"),
        "v1/v1.txt": "1.0",
        "v1/patches/patch.patch": patch_file.format(version="1.0"),

        "v2/conanfile.py": conanfile.format(version="v2"),
        "v2/conandata.yml": conandata_yml.format(version="2.0"),
        "v2/v2.txt": "2.0",
        "v2/patches/patch.patch": patch_file.format(version="2.0"),
    })

    tc.run("create v1 --version=1.0")
    v1_layout = tc.exported_layout()
    tc.test_v1_path = v1_layout.export()

    tc.run("create v2 --version=2.0")
    tc.test_v2_revision = tc.exported_layout().reference.revision

    tc.run("upload * -r=default -c")
    tc.run("remove * -c")

    return tc


@pytest.mark.parametrize("old_args", [
    "-op=v1 -or=pkg/1.0",
    "-or=pkg/1.0"
])
@pytest.mark.parametrize("new_args", [
    "-np=v2 -nr=pkg/2.0",
    "-nr=pkg/2.0#{revision}",
])
@pytest.mark.parametrize("formatter", [
    "-f=json --out-file=output.json",
    "-f=html --out-file=output.html"
])
@pytest.mark.tool("git")
def test_compare_paths(fixture_client, old_args, new_args, formatter):
    tc = fixture_client
    v2_revision = fixture_client.test_v2_revision
    v1_path = tc.test_v1_path.replace("\\", "/")

    new_args = new_args.format(revision=v2_revision)

    tc.run(f'report diff {old_args} {new_args} {formatter}')

    if "json" in formatter:
        output_json = json.loads(tc.load("output.json"))
        assert any(os.path.join(v1_path, "conanfile.py").replace("\\", "/") in k
                   for k in output_json.keys())
        assert any(os.path.join(v1_path, "conanmanifest.txt").replace("\\", "/") in k
                   for k in output_json.keys())
        # We have patch information
        assert any(os.path.abspath(os.path.join(v1_path, "..", "es", "patches", "patch.patch")).replace("\\", "/") in k
                   for k in output_json.keys())
        # We have exports information
        assert any(os.path.abspath(os.path.join(v1_path, "..", "e", "v1.txt")).replace("\\", "/") in k
                   for k in output_json.keys())
        # New files
        assert any("new-file-for-v2.txt" in k for k in output_json.keys())

    elif "html" in formatter:
        output_html = tc.load("output.html")
        # We have patch information
        assert "(old)/es/patches/patch.patch" in output_html
        assert "(new)/s/new-file-for-v2.txt" in output_html
        # We have exports information
        assert "(old)/e/v1.txt" in output_html



import os
import textwrap
import pytest

from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient
from conan.internal.util.files import save


complete_hook = """
import os

def pre_export(conanfile):
    conanfile.output.info("Hello")
    # TODO: To have the export_folder here needs a bit more deep refactoring
    assert conanfile.export_folder is None
    assert conanfile.recipe_folder, "recipe_folder not defined"

def post_export(conanfile):
    conanfile.output.info("Hello")
    assert conanfile.export_folder, "export_folder not defined"
    assert conanfile.export_sources_folder, "export_sources_folder not defined"
    assert conanfile.recipe_folder, "recipe_folder not defined"

def pre_validate(conanfile):
    conanfile.output.info("Hello")

def post_validate(conanfile):
    conanfile.output.info("Hello")

def pre_source(conanfile):
    conanfile.output.info("Hello")
    assert conanfile.source_folder, "source_folder not defined"

def post_source(conanfile):
    conanfile.output.info("Hello")
    assert conanfile.source_folder, "source_folder not defined"

def pre_generate(conanfile):
    conanfile.output.info("Hello")
    assert conanfile.generators_folder, "generators_folder not defined"

def post_generate(conanfile):
    conanfile.output.info("Hello")
    assert conanfile.generators_folder, "generators_folder not defined"

def pre_build(conanfile):
    conanfile.output.info("Hello")
    assert conanfile.source_folder, "source_folder not defined"
    assert conanfile.build_folder, "build_folder not defined"

def post_build(conanfile):
    conanfile.output.info("Hello")
    assert conanfile.source_folder, "source_folder not defined"
    assert conanfile.build_folder, "build_folder not defined"

def pre_package(conanfile):
    conanfile.output.info("Hello")
    assert conanfile.source_folder, "source_folder not defined"
    assert conanfile.build_folder, "build_folder not defined"
    assert conanfile.package_folder, "package_folder not defined"

def post_package(conanfile):
    conanfile.output.info("Hello")
    assert conanfile.source_folder, "source_folder not defined"
    assert conanfile.build_folder, "build_folder not defined"
    assert conanfile.package_folder, "package_folder not defined"

def pre_package_info(conanfile):
    conanfile.output.info("Hello")

def post_package_info(conanfile):
    conanfile.output.info("Hello")

def post_package_id(conanfile):
    conanfile.output.info("Hello")
"""


class TestHooks:

    def test_complete_hook(self):
        c = TestClient()
        hook_path = os.path.join(c.paths.hooks_path, "complete_hook", "hook_complete.py")
        save(hook_path, complete_hook)
        c.save({"conanfile.py": GenConanfile("pkg", "0.1")})

        c.run("source .")
        hook_msg = "[HOOK - complete_hook/hook_complete.py]"
        assert f"conanfile.py (pkg/0.1): {hook_msg} pre_source(): Hello" in c.out
        assert f"conanfile.py (pkg/0.1): {hook_msg} post_source(): Hello" in c.out

        c.run("install .")
        assert f"conanfile.py (pkg/0.1): {hook_msg} pre_validate(): Hello" in c.out
        assert f"conanfile.py (pkg/0.1): {hook_msg} post_validate(): Hello" in c.out
        assert f"conanfile.py (pkg/0.1): {hook_msg} pre_generate(): Hello" in c.out
        assert f"conanfile.py (pkg/0.1): {hook_msg} post_generate(): Hello" in c.out

        c.run("build .")
        assert f"conanfile.py (pkg/0.1): {hook_msg} pre_validate(): Hello" in c.out
        assert f"conanfile.py (pkg/0.1): {hook_msg} post_validate(): Hello" in c.out
        assert f"conanfile.py (pkg/0.1): {hook_msg} pre_build(): Hello" in c.out
        assert f"conanfile.py (pkg/0.1): {hook_msg} post_build(): Hello" in c.out
        assert f"conanfile.py (pkg/0.1): {hook_msg} pre_generate(): Hello" in c.out
        assert f"conanfile.py (pkg/0.1): {hook_msg} post_generate(): Hello" in c.out

        c.run("export . ")
        assert f"pkg/0.1: {hook_msg} pre_export(): Hello" in c.out
        assert f"pkg/0.1: {hook_msg} post_export(): Hello" in c.out

        c.run("export-pkg . ")
        assert f"pkg/0.1: {hook_msg} pre_export(): Hello" in c.out
        assert f"pkg/0.1: {hook_msg} post_export(): Hello" in c.out
        assert f"conanfile.py (pkg/0.1): {hook_msg} pre_package(): Hello" in c.out
        assert f"conanfile.py (pkg/0.1): {hook_msg} post_package(): Hello" in c.out
        assert f"conanfile.py (pkg/0.1): {hook_msg} post_package_id(): Hello" in c.out

        c.run("create . ")
        assert f"pkg/0.1: {hook_msg} pre_validate(): Hello" in c.out
        assert f"pkg/0.1: {hook_msg} post_validate(): Hello" in c.out
        assert f"pkg/0.1: {hook_msg} pre_export(): Hello" in c.out
        assert f"pkg/0.1: {hook_msg} post_export(): Hello" in c.out
        assert f"pkg/0.1: {hook_msg} pre_source(): Hello" in c.out
        assert f"pkg/0.1: {hook_msg} post_source(): Hello" in c.out
        assert f"pkg/0.1: {hook_msg} pre_build(): Hello" in c.out
        assert f"pkg/0.1: {hook_msg} post_build(): Hello" in c.out
        assert f"pkg/0.1: {hook_msg} pre_package(): Hello" in c.out
        assert f"pkg/0.1: {hook_msg} post_package(): Hello" in c.out
        assert f"pkg/0.1: {hook_msg} pre_package_info(): Hello" in c.out
        assert f"pkg/0.1: {hook_msg} post_package_info(): Hello" in c.out
        assert f"pkg/0.1: {hook_msg} post_package_id(): Hello" in c.out

    def test_import_hook(self):
        """ Test that a hook can import another random python file
        """
        custom_module = textwrap.dedent("""
            def my_printer(output):
                output.info("my_printer(): CUSTOM MODULE")
            """)

        my_hook = textwrap.dedent("""
            from custom_module.custom import my_printer

            def pre_export(conanfile):
                my_printer(conanfile.output)
            """)
        c = TestClient()
        hook_path = os.path.join(c.paths.hooks_path, "my_hook", "hook_my_hook.py")
        init_path = os.path.join(c.paths.hooks_path, "my_hook", "custom_module", "__init__.py")
        custom_path = os.path.join(c.paths.hooks_path, "my_hook", "custom_module", "custom.py")
        c.save({init_path: "",
                custom_path: custom_module,
                hook_path: my_hook,
                "conanfile.py": GenConanfile("pkg", "1.0")})

        c.run("export . ")
        assert "[HOOK - my_hook/hook_my_hook.py] pre_export(): my_printer(): CUSTOM MODULE" \
               in c.out

    def test_hook_raising(self):
        """ Test output when a hook raises
        """
        c = TestClient()
        my_hook = textwrap.dedent("""
            def pre_export(conanfile):
                raise Exception("Boom")
            """)
        hook_path = os.path.join(c.paths.hooks_path, "my_hook", "hook_my_hook.py")
        c.save({hook_path: my_hook,
                "conanfile.py": GenConanfile("pkg", "1.0")})

        c.run("export . ", assert_error=True)
        assert "ERROR: [HOOK - my_hook/hook_my_hook.py] pre_export(): Boom" in c.out

    def test_post_build_fail(self):
        """ Test the post_build_fail hook
        """
        c = TestClient()
        my_hook = textwrap.dedent("""
           def post_build_fail(conanfile):
               conanfile.output.info("Hello")
           """)
        hook_path = os.path.join(c.paths.hooks_path, "my_hook", "hook_my_hook.py")
        save(hook_path, my_hook)
        conanfile = textwrap.dedent("""
            from conan import ConanFile
            class Pkg(ConanFile):
                def build(self):
                    raise Exception("Boom!")
            """)
        c.save({"conanfile.py": conanfile})

        c.run("build . ", assert_error=True)
        assert "conanfile.py: [HOOK - my_hook/hook_my_hook.py] post_build_fail(): Hello" in c.out
        assert "ERROR: conanfile.py: Error in build() method, line 5" in c.out

    def test_validate_hook(self):
        """ The validate hooks are executed only if the method is declared in the recipe.
        """
        c = TestClient()
        hook_path = os.path.join(c.paths.hooks_path, "testing", "hook_complete.py")
        save(hook_path, complete_hook)

        conanfile = textwrap.dedent("""
            from conan import ConanFile
            class Pkg(ConanFile):
                def validate(self):
                    pass
            """)
        c.save({"conanfile.py": conanfile})

        c.run("create . --name=foo --version=0.1.0")
        assert f"foo/0.1.0: [HOOK - testing/hook_complete.py] pre_validate(): Hello" in c.out
        assert f"foo/0.1.0: [HOOK - testing/hook_complete.py] post_validate(): Hello" in c.out


@pytest.mark.parametrize("hook_name", ["pre_validate", "post_validate"])
def test_validate_invalid_configuration(hook_name):
    """ When raising ConanInvalidConfiguration in the pre_validate and post_validate hooks,
        it should be forwarded and preserve the same exception type.
    """
    hook = textwrap.dedent(f"""
        from conan.errors import ConanInvalidConfiguration
        def {hook_name}(conanfile):
            raise ConanInvalidConfiguration("Invalid configuration in {hook_name} hook")
    """)

    conanfile = textwrap.dedent(f"""
        from conan import ConanFile
        from conan.errors import ConanException

        class Pkg(ConanFile):
            def validate(self):
                if "{hook_name}" == "pre_validate":
                    raise ConanException("Should not reach this point")
    """)

    client = TestClient()
    hook_path = os.path.join(client.paths.hooks_path, "custom_hooks", "hook_pre_validate.py")
    client.save({"conanfile.py": conanfile,
                 hook_path: hook})

    client.run("build . ", assert_error=True)
    assert f"ERROR: conanfile.py: Invalid ID: Invalid: Invalid configuration in {hook_name} hook" in client.out
    assert "Should not reach this point" not in client.out



from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient


def test_test_requires():
    c = TestClient()
    c.save({"test/conanfile.py": GenConanfile("test", "0.1"),
            "consumer/conanfile.txt": "[test_requires]\ntest/0.1"})
    c.run("create test")
    c.run("install consumer")
    c.assert_listed_require({"test/0.1": "Cache"}, test=True)
    c.assert_listed_binary({"test/0.1": ("da39a3ee5e6b4b0d3255bfef95601890afd80709", "Cache")},
                           test=True)


def test_test_requires_options():
    c = TestClient()
    c.save({"test/conanfile.py": GenConanfile("test", "0.1").with_option("myoption", [1, 2, 3]),
            "consumer/conanfile.txt": "[test_requires]\ntest/0.1\n[options]\n*:myoption=2"})
    c.run("create test -o myoption=2")
    c.assert_listed_binary({"test/0.1": ("a3cb1345b8297bfdffea4ef4bb1b2694c54d1d69", "Build")})

    c.run("install consumer")
    c.assert_listed_require({"test/0.1": "Cache"}, test=True)
    c.assert_listed_binary({"test/0.1": ("a3cb1345b8297bfdffea4ef4bb1b2694c54d1d69", "Cache")},
                           test=True)
